#
# Class::Singleton test script
#
# Andy Wardley <abw@cre.canon.co.uk>
#

use Test::More
plan tests => 22

use Class::Singleton

ok 1

#========================================================================
#                         -- CLASS DEFINTIONS --
#========================================================================

#------------------------------------------------------------------------
# define 'DerivedSingleton', a class derived from Class::Singleton
#------------------------------------------------------------------------

package DerivedSingleton
use base 'Class::Singleton'


#------------------------------------------------------------------------
# define 'AnotherSingleton', a class derived from DerivedSingleton
#------------------------------------------------------------------------

package AnotherSingleton
use base 'DerivedSingleton'


#------------------------------------------------------------------------
# define 'ListSingleton', which uses a list reference as its type
#------------------------------------------------------------------------

package ListSingleton
use base 'Class::Singleton'

sub _new_instance
    my $class  = shift
    bless \$@, $class



#------------------------------------------------------------------------
# define 'ConfigSingleton', which has specific configuration needs.
#------------------------------------------------------------------------

package ConfigSingleton
use base 'Class::Singleton'

sub _new_instance
    my $class  = shift
    my $config = shift || \$%
    my $self = \%: 
        'one' => 'This is the first parameter'
        'two' => 'This is the second parameter'
        < $config->%
        
    bless $self, $class




#========================================================================
#                                -- TESTS --
#========================================================================

package main

# call Class::Singleton->instance() twice and expect to get the same
# reference returned on both occasions.

my $s1 = Class::Singleton->instance()

#2
ok(defined($s1), "   Class::Singleton instance 1: ")

my $s2 = Class::Singleton->instance()

#3
ok(defined($s2), "   Class::Singleton instance 2: ")

#4
ok($s1 \== $s2)


# call MySingleton->instance() twice and expect to get the same
# reference returned on both occasions.

my $s3 = DerivedSingleton->instance()

#5
ok(defined($s3))

my $s4 = DerivedSingleton->instance()

#6
ok(defined($s4))

#7
ok($s3 \== $s4)


# call MyOtherSingleton->instance() twice and expect to get the same
# reference returned on both occasions.

my $s5 = AnotherSingleton->instance()

#8
ok(defined($s5))

my $s6 = AnotherSingleton->instance()

#9
ok(defined($s6), "   AnotherSingleton instance 2: ")

#10
ok($s5 \== $s6, "AnotherSingleton instances are identical")


#------------------------------------------------------------------------
# having checked that each instance of the same class is the same, we now
# check that the instances of the separate classes are actually different
# from each other
#------------------------------------------------------------------------

#11-13
ok( $s1 \!= $s3 , "Class::Singleton and DerviedSingleton are different")
ok( $s1 \!= $s5 ,
    "Class::Singleton and AnotherSingleton are different")
ok( $s3 \!= $s5 ,
    "DerivedSingleton and AnotherSingleton are different")



#------------------------------------------------------------------------
# test ListSingleton
#------------------------------------------------------------------------

my $ls1 = ListSingleton->instance()
my $ls2 = ListSingleton->instance()

#14
ok( $ls1,
    "ListSingleton #1 is defined")

#15
ok( $ls2,
    "ListSingleton #2 is defined")

#16 - check they are the same reference
ok( $ls1 \== $ls2,
    "ListSingleton #1 and #2 correctly reference the same list")

#17 - check it's a LIST reference
ok( dump::view($ls1) =~ m/ARRAY/,
    "ListSingleton correctly contains a list reference")



#------------------------------------------------------------------------
# test ConfigSingleton
#------------------------------------------------------------------------

# create a ConfigSingleton
my $config = \%:  'foo' => 'This is foo' 
my $cs1 = ConfigSingleton->instance($config)

# add another parameter to the config
$config->{?'bar'} => 'This is bar'

# shouldn't call new() so changes to $config shouldn't matter
my $cs2 = ConfigSingleton->instance($config)

#18
ok( $cs1,
    "ConfigSingleton #1 is defined")

#19
ok( $cs2,
    "ConfigSingleton #2 is defined")

#20 - check they are the same reference
ok( $cs1 \== $cs2,
    "ConfigSingleton #1 and #2 correctly reference the same object")

#21 - check that 3 keys are defined in $cs1
ok( (nkeys $cs1->%) == 3,
    "ConfigSingleton #1 correctly has 3 keys")

#22 - and also in $cs2
ok( (nkeys $cs2->%) == 3,
    "ConfigSingleton #2 correctly has 3 keys")





