
package Compress::Zlib

require Exporter
use IO::Handle 
use Scalar::Util < qw(dualvar)

use Compress::Raw::Zlib v2.006 

use warnings 
use bytes 
our ($VERSION, @ISA, @EXPORT)

$VERSION = '2.007'
$VERSION = eval $VERSION

@ISA = qw(Exporter)
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw(
        compress uncompress
    )

push @EXPORT, < @Compress::Raw::Zlib::EXPORT 

BEGIN

    *zlib_version = &Compress::Raw::Zlib::zlib_version


# typeglob constants.
for my $name ((@:  <qw|DEF_WBITS MAX_MEM_LEVEL MAX_WBITS OS_CODE|
                   < grep { m/^Z_/ }, @Compress::Raw::Zlib::EXPORT))
    Symbol::fetch_glob($name)->* = Symbol::fetch_glob("Compress::Raw::Zlib::$name")->*->&


use constant FLAG_APPEND             => 1 
use constant FLAG_CRC                => 2 
use constant FLAG_ADLER              => 4 
use constant FLAG_CONSUME_INPUT      => 8 

sub compress
    my (@: $x, $output, $err, $in) =@: '', '', '', '' 

    if (ref @_[0] )
        $in = @_[0] 
        die "not a scalar reference" unless ref $in eq 'SCALAR' 
    else
        $in = \@_[0] 

    my $level = ((nelems @_) == 2 ?? @_[1] !! Z_DEFAULT_COMPRESSION() )

    $x = Compress::Raw::Zlib::Deflate->new( AppendOutput => 1, Level => $level)
    $x or return undef 

    $err = $x->deflate($in, $output) 
    return undef unless $err == Z_OK() 

    $err = $x->flush($output) 
    return undef unless $err == Z_OK() 

    return $output 



sub uncompress
    my (@: $x, $output, $err, $in) =@: '', '', '', '' 

    if (ref @_[0] )
        $in = @_[0] 
        die "not a scalar reference" unless ref $in eq 'SCALAR' 
    else
        $in = \@_[0] 
    

    $x = Compress::Raw::Zlib::Inflate->new( ConsumeInput => 0)
    $x or return undef 

    $err = $x->inflate($in, $output) 
    return undef unless $err == Z_STREAM_END() 

    return $output 


sub memGzip
    my $out

    # if the deflation buffer isn't a reference, make it one
    my $string = (ref @_[0] ?? @_[0] !! \@_[0]) 

    my $x = Compress::Raw::Zlib::Deflate->new( WindowBits => WANT_GZIP )
        or return undef 
    my $status = $x->deflate($string, $out)
    return undef
        unless $status == Z_OK()

    $status = $x->flush(my $finish_out)
    return undef unless $status == Z_OK()
    $out .= $finish_out

    return $out


sub memGunzip
    # if the buffer isn't a reference, make it one
    my $string = (ref @_[0] ?? @_[0] !! \@_[0])

    my $bufsize = length $string->$ +> 4096 ?? length $string->$ !! 4096 
    my $x =  Compress::Raw::Zlib::Inflate->new(\(%: WindowBits => WANT_GZIP
                                                    Bufsize => $bufsize))
    $x or return undef

    my $output = "" 
    my $status = $x->inflate($string, $output)
    return undef
        unless $status == Z_STREAM_END()

    if (length $string->$ +>= 8)
        my (@: $crc, $len) = @: unpack("VV", substr($string->$, 0, 8))
        substr($string->$, 0, 8, '')
        return undef
            unless $len == length($output) and
            $crc == crc32($output)
    else
        $string->$ = ''
    
    return $output


1
__END__


=head1 NAME

Compress::Zlib - Interface to zlib compression library

=head1 SYNOPSIS

    use Compress::Zlib ;

    $dest = compress($source) ;
    $dest = uncompress($source) ;

    $dest = Compress::Zlib::memGzip($buffer) ;
    $dest = Compress::Zlib::memGunzip($buffer) ;

    ZLIB_VERSION
    ZLIB_VERNUM



=head1 DESCRIPTION

The I<Compress::Zlib> module provides a Perl interface to the I<zlib>
compression library (see L</AUTHOR> for details about where to get
I<zlib>). 

The C<Compress::Zlib> module can be split into two general areas of
functionality, namely a simple read/write interface to I<gzip> files
and a low-level in-memory compression/decompression interface.

Each of these areas will be discussed in the following sections.

=head2 Compress::Zlib::memGzip

This function is used to create an in-memory gzip file with the minimum
possible gzip header (exactly 10 bytes).

    $dest = Compress::Zlib::memGzip($buffer) ;

If successful, it returns the in-memory gzip file, otherwise it returns
undef.

The C<$buffer> parameter can either be a scalar or a scalar reference.

See L<IO::Compress::Gzip|IO::Compress::Gzip> for an alternative way to
carry out in-memory gzip compression.

=head2 Compress::Zlib::memGunzip

This function is used to uncompress an in-memory gzip file.

    $dest = Compress::Zlib::memGunzip($buffer) ;

If successful, it returns the uncompressed gzip file, otherwise it
returns undef.

The C<$buffer> parameter can either be a scalar or a scalar reference. The
contents of the C<$buffer> parameter are destroyed after calling this function.

See L<IO::Uncompress::Gunzip|IO::Uncompress::Gunzip> for an alternative way
to carry out in-memory gzip uncompression.

=head1 COMPRESS/UNCOMPRESS

Two functions are provided to perform in-memory compression/uncompression of
RFC 1950 data streams. They are called C<compress> and C<uncompress>.

=over 5

=item B<$dest = compress($source [, $level] ) ;>

Compresses C<$source>. If successful it returns the compressed
data. Otherwise it returns I<undef>.

The source buffer, C<$source>, can either be a scalar or a scalar
reference.

The C<$level> parameter defines the compression level. Valid values are
0 through 9, C<Z_NO_COMPRESSION>, C<Z_BEST_SPEED>,
C<Z_BEST_COMPRESSION>, and C<Z_DEFAULT_COMPRESSION>.
If C<$level> is not specified C<Z_DEFAULT_COMPRESSION> will be used.


=item B<$dest = uncompress($source) ;>

Uncompresses C<$source>. If successful it returns the uncompressed
data. Otherwise it returns I<undef>.

The source buffer can either be a scalar or a scalar reference.

=back

Please note: the two functions defined above are I<not> compatible with
the Unix commands of the same name.

See L<IO::Deflate|IO::Deflate> and L<IO::Inflate|IO::Inflate> included with
this distribution for an alternative interface for reading/writing RFC 1950
files/buffers.

=head1 CONSTANTS

All the I<zlib> constants are automatically imported when you make use
of I<Compress::Zlib>.


=head1 SEE ALSO

L<IO::Compress::Gzip>, L<IO::Uncompress::Gunzip>, L<IO::Compress::Deflate>, L<IO::Uncompress::Inflate>, L<IO::Compress::RawDeflate>, L<IO::Uncompress::RawInflate>, L<IO::Compress::Bzip2>, L<IO::Uncompress::Bunzip2>, L<IO::Compress::Lzop>, L<IO::Uncompress::UnLzop>, L<IO::Compress::Lzf>, L<IO::Uncompress::UnLzf>, L<IO::Uncompress::AnyInflate>, L<IO::Uncompress::AnyUncompress>

L<Compress::Zlib::FAQ|Compress::Zlib::FAQ>

L<File::GlobMapper|File::GlobMapper>, L<Archive::Zip|Archive::Zip>,
L<Archive::Tar|Archive::Tar>,
L<IO::Zlib|IO::Zlib>


For RFC 1950, 1951 and 1952 see 
F<http://www.faqs.org/rfcs/rfc1950.html>,
F<http://www.faqs.org/rfcs/rfc1951.html> and
F<http://www.faqs.org/rfcs/rfc1952.html>

The I<zlib> compression library was written by Jean-loup Gailly
F<gzip@prep.ai.mit.edu> and Mark Adler F<madler@alumni.caltech.edu>.

The primary site for the I<zlib> compression library is
F<http://www.zlib.org>.

The primary site for gzip is F<http://www.gzip.org>.




=head1 AUTHOR

This module was written by Paul Marquess, F<pmqs@cpan.org>. 



=head1 MODIFICATION HISTORY

See the Changes file.

=head1 COPYRIGHT AND LICENSE

Copyright (c) 1995-2007 Paul Marquess. All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.



