
use warnings

use Config
use ExtUtils::MakeMaker
use File::Basename < qw(dirname)
use Getopt::Long

eval "use ExtUtils::MakeMaker::Coverage"
if( $^EVAL_ERROR )
    print $^STDOUT, "Skipping testcover target, ExtUtils::MakeMaker::Coverage not found\n"
else
    print $^STDOUT, "Adding testcover target\n"


our ($opt_default, $opt_libpath)

GetOptions(
    "default", \$opt_default,
    "lib=s",   \$opt_libpath,
    )

$opt_default ||= env::var('CRYPT_SSLEAY_DEFAULT')

# FIND POSSIBLE SSL INSTALLATIONS

my @POSSIBLE_SSL_DIRS
if ($opt_libpath)
    # explicit from command-line
    @POSSIBLE_SSL_DIRS = @: $opt_libpath
    $opt_default       = 1
elsif ($^OS_NAME eq 'MSWin32')
    @POSSIBLE_SSL_DIRS = @:  'c:\openssl' 
elsif ($^OS_NAME eq 'VMS')
    @POSSIBLE_SSL_DIRS = @:  '/ssl$root' 
else
    # Unix and the rest
    @POSSIBLE_SSL_DIRS = qw(
        /usr/local/openssl
        /usr/local/ssl
        /local/ssl
        /opt/ssl
        /usr/local
        /local
        /usr
    )


my @CANDIDATE = grep { $_ }, map {Candidate($_)}, @POSSIBLE_SSL_DIRS

if (nelems(@CANDIDATE) == 0)
    $opt_default = 0
    my $bar = '=' x 72
    print $^STDOUT, <<INFO
$bar
No installed SSL libraries found in any of the following places.
INFO
    for (@POSSIBLE_SSL_DIRS)
        print $^STDOUT, "    $_\n"
    print $^STDOUT, <<INFO
You will have to either specify a directory location at the following
prompt, or rerun the Makefile.PL program and use the --lib switch
to specify the path. If the path in question is considered standard
on your platform, please consider filing a bug report in order to
have it taken into account in a subsequent version of Crypt::SSLeay.

INFO


my $SSL_DIR
my $pkg_config

if (nelems(@CANDIDATE) == 1)
    $pkg_config = @CANDIDATE[0]
    if ($opt_default)
        $SSL_DIR = $pkg_config->{?dir}
    else
        print $^STDOUT, <<INFO
=======================================================
Only one $pkg_config->{?type} installation found at $pkg_config->{?dir}
Consider running 'perl Makefile.PL --default' the next
time Crypt::SSLeay is upgraded to select this directory
automatically thereby avoiding the following prompt.
=======================================================
INFO
    
elsif (nelems(@CANDIDATE) +> 1)
    print $^STDOUT, "Found multiple possibilities for OpenSSL\n"
    for my $c ( @CANDIDATE)
        print $^STDOUT, "  $c->{?dir} ($c->{?type} $c->{?ver})\n"
    


if (not $SSL_DIR)
    my %cand
    for my $c ( @CANDIDATE)
        %cand{+$c->{?dir}} = \%: < $c->%
    

    $SSL_DIR = prompt "Which SSL install path do you want to use?",
                      @CANDIDATE[0]->{?dir}

    if (exists %cand{$SSL_DIR})
        # we've already determined that this directory is usable
        $pkg_config = %cand{?$SSL_DIR}
    else
        # unknown directory, better check it out
        $pkg_config = Candidate($SSL_DIR)
        if (not $pkg_config)
            die <<INFO
$SSL_DIR does not appear to be an SSL library installation, since
the required header files were not found. The build cannot proceed.
INFO
        
    


# note: $SSL_DIR is now sane at this point

my (@INC_FLAGS, @LIB_FLAGS)

if ($^OS_NAME eq 'VMS')
    push @INC_FLAGS, "-I$pkg_config->{?inc}"
    push @LIB_FLAGS, < qw(-L/SYS$SHARE -lSSL$LIBSSL_SHR32 -lSSL$LIBCRYPTO_SHR32)
elsif ($^OS_NAME eq 'MSWin32')
    # external tools probably expect \ and not / for path separators
    $SSL_DIR =~ s{/}{\\}g

    # default to drive C:
    $SSL_DIR = "c:$SSL_DIR" if $SSL_DIR !~ m/\A[a-z]:/i

    my $inc = $pkg_config->{?inc}
    $inc =~ s{/}{\\}g
    $inc !~ m/^[a-z]:/i and $inc = "c:$inc"
    push @INC_FLAGS, "-I$inc"
    push @INC_FLAGS, "-I$SSL_DIR\\inc32" if -d "$SSL_DIR/inc32"

    my $vanilla = config_value('cc') eq 'gcc' ?? 1 !! 0
    $vanilla and print $^STDOUT, "Assuming Vanilla/Strawberry Perl installation\n"

    if ($vanilla and -d "$SSL_DIR\\lib\\MinGW")
        push @LIB_FLAGS, "-L$SSL_DIR\\lib\\MinGW"
    elsif(-d "$SSL_DIR/lib")
        push @LIB_FLAGS, "-L$SSL_DIR\\lib"
    elsif(-d "$SSL_DIR/out32dll")
        # patch from Ben Laurie
        push @LIB_FLAGS, "-L$SSL_DIR\\out32dll"
    else
        # Allow developers to point at OpenSSL source...
        push @LIB_FLAGS, "-L$SSL_DIR"
    

    push @LIB_FLAGS, < qw(-lssleay32 -llibeay32)
    push @LIB_FLAGS, < qw(-lRSAglue -lrsaref) if $pkg_config->{?type} ne 'OpenSSL'
else
    push @INC_FLAGS, "-I$pkg_config->{?inc}"

    push @LIB_FLAGS, "-L$SSL_DIR/lib", < qw(-lssl -lcrypto -lgcc)
    push @LIB_FLAGS, < qw(-lRSAglue -lrsaref) if $pkg_config->{?type} ne 'OpenSSL'

    # ccc on alpha support
    if ($^OS_NAME eq 'linux'
          and `uname -m` =~ m/alpha/
          and !(system("nm $SSL_DIR/lib/libssl.a|grep -q 'U _Ots'")>>8)
        )
        push @LIB_FLAGS, '-lots'
    

    # this fix was suggested for building on RedHat 9
    push @INC_FLAGS, '-I/usr/kerberos/include' if -d '/usr/kerberos/include'


# write include file that determines ssl support
my $ssl_header_prefix = ($pkg_config->{?inc} =~ m/\bopenssl/i)
    ?? 'openssl/'
    !! ''


# we need to include crypto.h for SSLeay so the version gets picked up in SSLeay.xs

open(my $include_fh, ">", "crypt_ssleay_version.h") || die("can't open crypt_ssleay_version.h for writing: $^OS_ERROR")
print $include_fh, <<INFO
#include "$($ssl_header_prefix)ssl.h"
#include "$($ssl_header_prefix)crypto.h"
#include "$($ssl_header_prefix)err.h"
#include "$($ssl_header_prefix)rand.h"
#include "$($ssl_header_prefix)pkcs12.h"

INFO

# OPENSSL_free defined in OpenSSL 0.9.6 and higher
my $openssl_version = $pkg_config->{?ver}
if(($openssl_version =~ m/^OpenSSL/) and ($openssl_version !~ m/0\.9\.[2-5]/))
    print $include_fh, "#define CRYPT_SSLEAY_free OPENSSL_free\n"
else
    print $include_fh, "#define CRYPT_SSLEAY_free free\n"


close $include_fh or die "Cannot close crypt_ssleay_version.h for output: $^OS_ERROR\n"

print $^STDOUT, <<INFO

BUILD INFORMATION
================================================
ssl library: $pkg_config->{?type} $pkg_config->{?ver} in $SSL_DIR
ssl header:  $($ssl_header_prefix)ssl.h
libraries:   $(join ' ',@LIB_FLAGS)
include dir: $(join ' ',@INC_FLAGS)
================================================
INFO

WriteMakefile(
    NAME          => 'Crypt::SSLeay',
    AUTHOR        => 'David Landgren',
    ABSTRACT_FROM => 'SSLeay.pm',
    VERSION_FROM  => 'SSLeay.pm',
    LICENSE       => 'perl',
    LIBS          => (@: "$(join ' ',@LIB_FLAGS)"),
    INC           => "$(join ' ',@INC_FLAGS)",
    NEEDS_LINKING => 1,
    clean         => (%: 
    FILES => 'crypt_ssleay_version.h test.config'
    ),
    )

if (open my $out_fh, ">", 'test.config')
    print $out_fh, <<INFO
ssl $pkg_config->{?type} $pkg_config->{?ver} in $SSL_DIR
lib $(join ' ',@LIB_FLAGS)
inc $(join ' ',@INC_FLAGS)
cc $(config_value('cc'))
INFO

    print $^STDOUT, <<INFO
The test suite can attempt to connect to public servers
to ensure that the code is working properly. If you are
behind a strict firewall or have no network connectivity,
these tests may fail (through no fault of the code).

INFO
    my $network_tests = prompt
        "Do you want to run the live tests (y/N) ?",
        'N'
    print $out_fh, "network_tests ", ($network_tests =~ m/y/i) ?? 1 !! 0, "\n"
    close $out_fh


## HELPERS

sub Candidate
    my $dir = shift
    return unless -d $dir

    my $inc_dir
    my $version_file
    for (@: 
         "$dir/crypto/opensslv.h" # cygwin32 builds
         "$dir/inc32/openssl/opensslv.h" # old win32 builds
         "$dir/include/openssl/opensslv.h"
         "$dir/include/opensslv.h"
         "$dir/include/crypto.h"
        )
        if (-e $_)
            $version_file = $_
            last
        
    
    return unless defined $version_file

    my $open_ssl = ($version_file =~ m/openssl/i) ?? 1 !! 0
    $inc_dir = dirname($version_file)
    return unless -e "$inc_dir/ssl.h"

    if ($^OS_NAME eq 'MSWin32')
        $inc_dir =~ s{[\\/]openssl\z}{}
    

    open(my $version_fh, "<", $version_file) or return
    my $version_match = $open_ssl ?? "OPENSSL_VERSION_NUMBER" !! "SSLEAY_VERSION_NUMBER"
    my $version
    my $type
    while ( ~< $version_fh->*)
        if (m/^#define\s+$version_match\s+0x0+(\d\d\d)/)
            $version = $1
            $version =~ s/(\d)0(\d)/$1$2/
            $type = ($version +> 92) ?? "OpenSSL" !! "SSLeay"
            $version = join('.', split(m//, "0$version"))
            last
        
    
    close($version_fh)

    # Silly test to look for the library files
    my $foundlib = 0
    my $libd
    if (-d "$dir/out32dll")
        $libd = 'out32dll'
    elsif ($^OS_NAME eq 'MSWin32' and config_value('cc') eq 'gcc')
        $libd = 'lib/MinGW'
    else
        $libd = 'lib'
    
    if (opendir(my $libdh, "$dir/$libd"))
        while (defined($_ = readdir($libdh)))
            if (m/\A(?:lib(?:crypto|eay32|ssl)|ssleay32)/)
                $foundlib++
                last
            
        
        closedir($libdh)
    
    print $^STDOUT, "$dir/$libd does not seem to contain the SSLeay library files\n"
        unless $foundlib

    return \%: 
        dir  => $dir
        inc  => $inc_dir
        ver  => $version
        type => $type
        

