package DBD::Gofer::Transport::pipeone

#   $Id: pipeone.pm 9560 2007-05-13 15:45:04Z timbo $
#
#   Copyright (c) 2007, Tim Bunce, Ireland
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.

use warnings

use Fcntl
use IO::Select
use IPC::Open3 < qw(open3)
use Symbol < qw(gensym)

use base < qw(DBD::Gofer::Transport::Base)

our $VERSION = sprintf("0.\%06d", (@: q$Revision: 9560 $ =~ m/(\d+)/)[0])

__PACKAGE__->mk_accessors( <qw(
    connection_info
    go_perl
))


sub new($self, $args)
    $args->{+go_perl} ||= do
        (\(@:  $^EXECUTABLE_NAME )) # FIXME blib

    if (not ref $args->{?go_perl})
        # user can override the perl to be used, either with an array ref
        # containing the command name and args to use, or with a string
        # (ie via the DSN) in which case, to enable args to be passed,
        # we split on two or more consecutive spaces (otherwise the path
        # to perl couldn't contain a space itself).
        $args->{+go_perl} = \$: split m/\s{2,}/, $args->{?go_perl}

    return $self->SUPER::new($args)



# nonblock($fh) puts filehandle into nonblocking mode
sub nonblock
    my $fh = shift
    my $flags = fcntl($fh, F_GETFL, 0)
        or die "Can't get flags for filehandle $fh: $^OS_ERROR"
    fcntl($fh, F_SETFL, $flags ^|^ O_NONBLOCK)
        or die "Can't make filehandle $fh nonblocking: $^OS_ERROR"



sub start_pipe_command($self, $cmd)
    $cmd = \(@:  $cmd ) unless ref $cmd eq 'ARRAY'

    # if it's important that the subprocess uses the same
    # (versions of) modules as us then the caller should
    # set PERL5LIB itself.

    # limit various forms of insanity, for now
    local env::var('DBI_TRACE') = undef # use DBI_GOFER_TRACE instead
    local env::var('DBI_PROFILE') = undef

    my (@: $wfh, $rfh, $efh) = @: gensym, gensym, gensym
    my $pid = open3($wfh, $rfh, $efh, < $cmd->@)
        or die "error starting $(join ' ', $cmd->@): $^OS_ERROR\n"
    if ($self->trace)
        $self->trace_msg(sprintf("Started pid $pid: $(join ' ', $cmd->@) \{fd: w\%d r\%d e\%d, ppid=$^PID\}\n", fileno $wfh, fileno $rfh, fileno $efh),0)

    nonblock($rfh)
    nonblock($efh)
    my $ios = IO::Select->new($rfh, $efh)

    return \%:
        cmd=>$cmd
        pid=>$pid
        wfh=>$wfh, rfh=>$rfh, efh=>$efh
        ios=>$ios




sub cmd_as_string
    my $self = shift
    # XXX meant to return a properly shell-escaped string suitable for system
    # but its only for debugging so that can wait
    my $connection_info = $self->connection_info
    return join " ", map { (m/^[-:\w]*$/) ?? $_ !! "'$_'" }, $connection_info->{?cmd}->@



sub transmit_request_by_transport($self, $request)

    my $frozen_request = $self->freeze_request($request)

    my $cmd = \@:  < $self->go_perl->@, < qw(-MDBI::Gofer::Transport::pipeone -e run_one_stdio)
    my $info = $self->start_pipe_command($cmd)

    my $wfh = delete $info->{wfh}
    # send frozen request
    print $wfh, $frozen_request
        or warn "error writing to $(join ' ', $cmd->@): $^OS_ERROR\n"
    # indicate that there's no more
    close $wfh
        or die "error closing pipe to $(join ' ', $cmd->@): $^OS_ERROR\n"

    $self->connection_info( $info )
    return



sub read_response_from_fh($self, $fh_actions)
    my $trace = $self->trace

    my $info = $self->connection_info || die
    my (@: $ios) =  $info->{[qw(ios)]}
    my $errors = 0
    my $complete

    die "No handles to read response from" unless $ios->count

    while ($ios->count)
        my @readable = $ios->can_read()
        for my $fh ( @readable)
            local $_ = undef
            my $actions = $fh_actions->{?dump::view($fh)} || die "panic: no action for $fh"
            my $rv = sysread($fh, ($_=''), 1024*31)  # to fit in 32KB slab
            unless ($rv)              # error (undef) or end of file (0)
                my $action
                unless (defined $rv)  # was an error
                    $self->trace_msg("error on handle $fh: $^OS_ERROR\n") if $trace +>= 4
                    $action = $actions->{?error} || $actions->{?eof}
                    ++$errors
                # XXX an error may be a permenent condition of the handle
                # if so we'll loop here - not good
                else
                    $action = $actions->{?eof}
                    $self->trace_msg("eof on handle $fh\n") if $trace +>= 4

                if ($action->($fh))
                    $self->trace_msg("removing $fh from handle set\n") if $trace +>= 4
                    $ios->remove($fh)

                next

            # action returns true if the response is now complete
            # (we finish all handles
            $actions->{?read}->($fh) && ++$complete

        last if $complete

    return $errors



sub receive_response_by_transport
    my $self = shift

    my $info = $self->connection_info || die
    my (@: $pid, $rfh, $efh, $ios, $cmd) =  $info->{[qw(pid rfh efh ios cmd)]}

    my $frozen_response
    my $stderr_msg

    $self->read_response_from_fh( \(%:
        dump::view($efh) => \ %:
            error => sub (@< @_) { warn "error reading response stderr: $^OS_ERROR"; 1 }
            eof   => sub (@< @_) { warn "eof on stderr" if 0; 1 }
            read  => sub (@< @_) { $stderr_msg .= $_; 0 }
        dump::view($rfh) => \ %:
            error => sub (@< @_) { warn "error reading response: $^OS_ERROR"; 1 }
            eof   => sub (@< @_) { warn "eof on stdout" if 0; 1 }
            read  => sub (@< @_) { $frozen_response .= $_; 0 }
        ))

    waitpid $info->{?pid}, 0
        or warn "waitpid: $^OS_ERROR" # XXX do something more useful?

    die ref($self)." command ($(join ' ', $cmd->@)) failed: $stderr_msg"
        if not $frozen_response # no output on stdout at all

    # XXX need to be able to detect and deal with corruption
    my $response = $self->thaw_response($frozen_response)

    if ($stderr_msg)
        # add stderr messages as warnings (for PrintWarn)
        $response->add_err(0, $stderr_msg, undef, $self->trace)


    return $response



1

__END__

=head1 NAME

DBD::Gofer::Transport::pipeone - DBD::Gofer client transport for testing

=head1 SYNOPSIS

  $original_dsn = "...";
  DBI->connect("dbi:Gofer:transport=pipeone;dsn=$original_dsn",...)

=head1 DESCRIPTION

Connect via DBD::Gofer and execute each request by starting executing a subprocess.

This is, as you might imagine, spectacularly inefficient!

It's only intended for testing. Specifically it demonstrates that the server
side is completely stateless.

It also provides a base class for the much more useful L<DBD::Gofer::Transport::stream>
transport.

=head1 AUTHOR

Tim Bunce, L<http://www.linkedin.com/in/timbunce>

=head1 LICENCE AND COPYRIGHT

Copyright (c) 2007, Tim Bunce, Ireland. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.

=head1 SEE ALSO

L<DBD::Gofer::Transport::Base>

L<DBD::Gofer>

=cut
