#!perl -w

use lib < qw(blib/arch blib/lib)	# needed since -T ignores PERL5LIB
use DBI < qw(:sql_types)
use Config
use Cwd

$^WARNING = 1
$^OUTPUT_AUTOFLUSH = 1

my $haveFileSpec = try { require File::Spec }
require VMS::Filespec if $^OS_NAME eq 'VMS'

use Test::More tests => 184

# "globals"
my ($r, $dbh)

ok !try {
    $dbh = DBI->connect("dbi:NoneSuch:foobar", 1, 1, \(%:  RaiseError => 1, AutoCommit => 1 ));
}, 'connect should fail'
like($^EVAL_ERROR->{?description}, qr/install_driver\(NoneSuch\) failed/, '... we should have an exception here')
ok(!$dbh, '... $dbh2 should not be defined')

$dbh = DBI->connect('dbi:ExampleP:', '', '')

sub check_connect_cached
    # connect_cached
    # ------------------------------------------
    # This test checks that connect_cached works
    # and how it then relates to the CachedKids
    # attribute for the driver.

    ok(my $dbh_cached_1 = DBI->connect_cached('dbi:ExampleP:', '', '', \(%:  TraceLevel=>0, Executed => 0 )))

    ok(my $dbh_cached_2 = DBI->connect_cached('dbi:ExampleP:', '', '', \(%:  TraceLevel=>0, Executed => 0 )))

    is($dbh_cached_1, $dbh_cached_2, '... these 2 handles are cached, so they are the same')

    ok(my $dbh_cached_3 = DBI->connect_cached('dbi:ExampleP:', '', '', \(%:  examplep_foo => 1 )))

    isnt($dbh_cached_3, $dbh_cached_2, '... this handle was created with different parameters, so it is not the same')

    # check that cached_connect applies attributes to handles returned from the cache
    # (The specific case of Executed is relevant to DBD::Gofer retry-on-error logic)
    ok $dbh_cached_1->do("select * from .") # set Executed flag
    ok $dbh_cached_1->Executed, 'Executed should be true'
    ok(my $dbh_cached_4 = DBI->connect_cached('dbi:ExampleP:', '', '', \(%:  TraceLevel=>0, Executed => 0 )))
    is $dbh_cached_4, $dbh_cached_1, 'should return same handle'
    ok !$dbh_cached_4->Executed, 'Executed should be false because reset by connect attributes'

    my $drh = $dbh->Driver
    isa_ok($drh, "DBI::dr")

    my @cached_kids = values $drh->CachedKids->% 
    ok(eq_set(\@cached_kids, \(@:  $dbh_cached_1, $dbh_cached_3 )), '... these are our cached kids')


check_connect_cached()

$dbh->STORE("AutoCommit" => 1)
$dbh->STORE("PrintError" => 0)

ok($dbh->AutoCommit == 1)
cmp_ok($dbh->PrintError, '==', 0, '... PrintError should be 0')

diag "others\n"
try { $dbh->commit('dummy') }
like($^EVAL_ERROR->{?description}, qr/DBI commit: invalid number of arguments:/)
    unless $DBI::PurePerl && ok(1)

ok($dbh->ping, "ping should return true")

# --- errors
my $cursor_e = $dbh->prepare("select unknown_field_name from ?")
is($cursor_e, undef, "prepare should fail")
ok($dbh->err, "sth->err should be true")
ok(DBI::err, "DBI::err should be true")
cmp_ok(DBI::err,    'eq', $dbh->err   , "DBI::err should match \$dbh->err")
like(DBI::errstr, qr/Unknown field names: unknown_field_name/, "DBI::errstr should contain error string")
cmp_ok(DBI::errstr, 'eq', $dbh->errstr, "DBI::errstr should match \$dbh->errstr")

# --- func
ok($dbh->errstr eq $dbh->func('errstr'))

my $std_sql = "select mode,size,name from ?"
my $csr_a = $dbh->prepare($std_sql)
ok(ref $csr_a)
ok($csr_a->NUM_OF_FIELDS == 3)

SKIP: do
    skip "inner/outer handles not fully supported for DBI::PurePerl", 2 if $DBI::PurePerl
    is($csr_a->Database, $dbh, '$csr_a->{Database} ne $dbh')
        unless $dbh->$->{?mx_handle_list} && ok(1) # skip for Multiplex tests
    ok($csr_a->Database->Driver )	# ie is 'outer' handle


my $driver_name = $csr_a->Database->Driver->Name
ok($driver_name eq 'ExampleP')

# --- FetchHashKeyName
my $csr_b = $dbh->prepare($std_sql)
$csr_b->execute('.')
ok(ref $csr_b)

ok($csr_a \!= $csr_b)

ok("$(join ' ', $csr_b->NAME)"    eq "mode size name")

# get a dir always readable on all platforms
my $dir = getcwd() || cwd()
$dir = VMS::Filespec::unixify($dir) if $^OS_NAME eq 'VMS'
# untaint $dir
#$dir =~ m/(.*)/; $dir = $1 || die;


# ---

my($col0, $col1, $col2, $col3, $rows)
my(@row_a, @row_b)

ok($csr_a->bind_columns(undef, \$col0, \$col1, \$col2) )
ok($csr_a->execute( $dir ), DBI::errstr)

@row_a = $csr_a->fetchrow_array 
ok(nelems(@row_a))

# check bind_columns
is(@row_a[0], $col0)
is(@row_a[1], $col1)
is(@row_a[2], $col2)

ok( ! $csr_a->bind_columns(undef, \$col0, \$col1) )
like $csr_a->errstr, '/bind_columns called with 2 values but 3 are needed/', 'errstr should contain error message'
ok( ! $csr_a->bind_columns(undef, \$col0, \$col1, \$col2, \$col3) )
like $csr_a->errstr, '/bind_columns called with 4 values but 3 are needed/', 'errstr should contain error message'

ok( $csr_a->bind_col(2, undef, \(%:  foo => 42 )) )
ok ! try { $csr_a->bind_col(0, undef) }
like $^EVAL_ERROR->{?description}, '/bind_col: column 0 is not a valid column \(1..3\)/', 'errstr should contain error message'
ok ! try { $csr_a->bind_col(4, undef) }
like $^EVAL_ERROR->{?description}, '/bind_col: column 4 is not a valid column \(1..3\)/', 'errstr should contain error message'

ok($csr_b->bind_param(1, $dir))
ok($csr_b->execute())
@row_b =  $csr_b->fetchrow_arrayref->@ 
ok(nelems(@row_b))

ok("$(join ' ', @row_a)" eq "$(join ' ', @row_b)")
@row_b = $csr_b->fetchrow_array 
ok("$(join ' ', @row_a)" ne "$(join ' ', @row_b)")

ok($csr_a->finish)
ok($csr_b->finish)

$csr_a = undef  # force destruction of this cursor now
ok(1)

diag "fetchrow_hashref('NAME')\n"
ok($csr_b->execute())
my $row_b = $csr_b->fetchrow_hashref()
ok($row_b)
ok($row_b->{?mode} == @row_a[0])
ok($row_b->{?size} == @row_a[1])
ok($row_b->{?name} eq @row_a[2])

diag "FetchHashKeyName\n"
ok($csr_b->execute())
$row_b = $csr_b->fetchrow_hashref()
ok($row_b)
ok(nelems(keys($row_b->%)) == 3)
ok($row_b->{?mode} == @row_a[0])
ok($row_b->{?size} == @row_a[1])
ok($row_b->{?name} eq @row_a[2])

diag "fetchall_arrayref\n"
ok($csr_b->execute())
$r = $csr_b->fetchall_arrayref
ok($r)
ok(nelems($r->@))
ok($r->[0]->[0] == @row_a[0])
ok($r->[0]->[1] == @row_a[1])
ok($r->[0]->[2] eq @row_a[2])

diag "fetchall_arrayref array slice\n"
ok($csr_b->execute())
$r = $csr_b->fetchall_arrayref(\(@: 2,1))
ok($r && nelems($r->@))
ok($r->[0]->[1] == @row_a[1])
ok($r->[0]->[0] eq @row_a[2])

diag "fetchall_arrayref hash slice\n"
ok($csr_b->execute())
$r = $csr_b->fetchall_arrayref(\(%:  SizE=>1, nAMe=>1))
ok($r && nelems($r->@))
ok($r->[0]->{?SizE} == @row_a[1])
ok($r->[0]->{?nAMe} eq @row_a[2])

diag "fetchall_arrayref hash\n"
ok($csr_b->execute())
$r = $csr_b->fetchall_arrayref(\(%: ))
ok($r)
ok(nelems(keys $r->[0]->%) == 3)
is((join ' ', $r->[0]->{[qw(mode size name)]}), "$(join ' ', @row_a)")

# use Data::Dumper; warn Dumper([\@row_a, $r]);

$rows = $csr_b->rows
ok($rows +> 0, "row count $rows")
ok($rows == nelems($r->@), "$rows vs ".nelems($r->@))
ok($rows == DBI::rows, "$rows vs DBI::rows")

# ---

diag "selectrow_array\n"
@row_b = $dbh->selectrow_array($std_sql, undef, $dir)
ok(nelems(@row_b) == 3)
ok("$(join ' ', @row_b)" eq "$(join ' ', @row_a)")

diag "selectrow_hashref\n"
$r = $dbh->selectrow_hashref($std_sql, undef, $dir)
is_deeply( $r, \(%:  mode => @row_a[0], size => @row_a[1], name => @row_a[2] ) )

diag "selectall_arrayref\n"
$r = $dbh->selectall_arrayref($std_sql, undef, $dir)
ok($r)
ok(nelems($r->[0]->@) == 3)
ok("$(join ' ', $r->[0]->@)" eq "$(join ' ', @row_a)")
ok(nelems($r->@) == $rows)

diag "selectall_arrayref Slice array slice\n"
$r = $dbh->selectall_arrayref($std_sql, \(%:  Slice => \(@:  2, 0 ) ), $dir)
ok($r)
ok(nelems($r->[0]->@) == 2)
ok("$(join ' ', $r->[0]->@)" eq "@row_a[2] @row_a[0]", qq{"$(join ' ', $r->[0]->@)" eq "@row_a[2] @row_a[0]"})
ok(nelems($r->@) == $rows)

diag "selectall_arrayref Columns array slice\n"
$r = $dbh->selectall_arrayref($std_sql, \(%:  Columns => \(@:  3, 1 ) ), $dir)
ok($r)
ok(nelems($r->[0]->@) == 2)
ok("$(join ' ', $r->[0]->@)" eq "@row_a[2] @row_a[0]", qq{"$(join ' ', $r->[0]->@)" eq "@row_a[2] @row_a[0]"})
ok(nelems($r->@) == $rows)

diag "selectall_arrayref hash slice\n"
$r = $dbh->selectall_arrayref($std_sql, \(%:  Columns => \(%:  MoDe=>1, NamE=>1 ) ), $dir)
ok($r)
is_deeply($r->[0], \(%:  MoDe => @row_a[0], NamE => @row_a[2] ) )
ok(nelems($r->@) == $rows)

$r = $dbh->selectall_hashref($std_sql, 'name', undef, $dir)
ok($r, "selectall_hashref result")
is(ref $r, 'HASH', "selectall_hashref HASH: ".ref $r)
is(nelems( keys $r->% ), $rows)
is($r->{ @row_a[2] }->{?size}, @row_a[1], qq{$r->{ @row_a[2] }->{?size} eq @row_a[1]})

diag "selectall_hashref by column number\n"
$r = $dbh->selectall_hashref($std_sql, 3, undef, $dir)
ok($r)
ok($r->{ @row_a[2] }->{?size} eq @row_a[1], qq{$r->{ @row_a[2] }->{?size} eq @row_a[1]})

diag "selectcol_arrayref\n"
$r = $dbh->selectcol_arrayref($std_sql, undef, $dir)
ok($r)
ok(nelems($r->@) == $rows)
ok($r->[0] eq @row_b[0])

diag "selectcol_arrayref column slice\n"
$r = $dbh->selectcol_arrayref($std_sql, \(%:  Columns => \(@: 3,2) ), $dir)
ok($r)
# use Data::Dumper; warn Dumper([\@row_b, $r]);
ok(nelems($r->@) == $rows * 2)
ok($r->[0] eq @row_b[2])
ok($r->[1] eq @row_b[1])

# ---

diag "others...\n"
my $csr_c
$csr_c = $dbh->prepare("select unknown_field_name1 from ?")
ok(!defined $csr_c)
ok(DBI::errstr =~ m/Unknown field names: unknown_field_name1/)

diag "RaiseError & PrintError & ShowErrorStatement\n"
$dbh->STORE("RaiseError" => 1)
ok($dbh->RaiseError)
$dbh->STORE("ShowErrorStatement" => 1)
ok($dbh->ShowErrorStatement)

my $error_sql = "select unknown_field_name2 from ?"

ok(! try { $csr_c = $dbh->prepare($error_sql); 1; })
#diag "$@\n";
like($^EVAL_ERROR->{?description}, qr/\Q$error_sql/) # ShowErrorStatement
like($^EVAL_ERROR->{?description}, qr/.*Unknown field names: unknown_field_name2/)

my $se_sth1 = $dbh->prepare("select mode from ?")
ok($se_sth1->RaiseError)
ok($se_sth1->ShowErrorStatement)

# check that $dbh->{Statement} tracks last _executed_ sth
ok($se_sth1->Statement eq "select mode from ?")
ok($dbh->Statement     eq "select mode from ?") or diag "got: $dbh->{?Statement}\n"
my $se_sth2 = $dbh->prepare("select name from ?")
ok($se_sth2->Statement eq "select name from ?")
ok($dbh->Statement     eq "select name from ?")
$se_sth1->execute('.')
ok($dbh->Statement     eq "select mode from ?")

# show error param values
ok(! try { $se_sth1->execute('first','second') })       # too many params
like($^EVAL_ERROR->{?description}, qr/\b1='first'/)
like($^EVAL_ERROR->{?description}, qr/\b2='second'/)

$se_sth1->finish
$se_sth2->finish

$dbh->setRaiseError(0)
ok(!$dbh->RaiseError)
$dbh->setShowErrorStatement(0)
ok(!$dbh->ShowErrorStatement)

do
    my @warn
    local($^WARN_HOOK) = sub (@< @_) { push @warn, @_[0]->message }
    $dbh->setPrintError(1)
    ok($dbh->PrintError)
    ok(! $dbh->selectall_arrayref("select unknown_field_name3 from ?"))
    ok("$(join ' ', @warn)" =~ m/Unknown field names: unknown_field_name3/)
    $dbh->setPrintError(0)
    ok(!$dbh->PrintError)



diag "HandleError\n"
my $HandleErrorReturn
my $HandleError = sub { do
        my $msg = sprintf "HandleError: \%s [h=\%s, rv=\%s, #=\%d]",
            < map { dump::view($_) }, @:  @_[0],@_[1],(defined(@_[2])??@_[2]!!'undef'),scalar(nelems(@_))
        die $msg   if $HandleErrorReturn +< 0
        diag "$msg\n"
        @_[2] = 42 if $HandleErrorReturn == 2
        return $HandleErrorReturn
    }


$dbh->setHandleError($HandleError)
ok($dbh->HandleError)
ok($dbh->HandleError &== $HandleError)

$dbh->setRaiseError(1)
$dbh->setPrintError(0)
$error_sql = "select unknown_field_name2 from ?"

diag "HandleError -> die\n"
$HandleErrorReturn = -1
dies_like( sub (@< @_) { $csr_c = $dbh->prepare($error_sql); 1; },
           qr/^HandleError:/ )

diag "HandleError -> 0 -> RaiseError\n"
$HandleErrorReturn = 0
ok(! try { $csr_c = $dbh->prepare($error_sql); 1; })
like($^EVAL_ERROR->message, qr/^DBD::(ExampleP|Multiplex|Gofer)::db prepare failed:/)

diag "HandleError -> 1 -> return (original)undef\n"
$HandleErrorReturn = 1
$r = try { $csr_c = $dbh->prepare($error_sql); }
ok(!$^EVAL_ERROR, $^EVAL_ERROR)
ok(!defined($r), $r)

diag "HandleError -> 2 -> return (modified)42\n"
$HandleErrorReturn = 2
$r = try { $csr_c = $dbh->prepare($error_sql) }
ok(!$^EVAL_ERROR, $^EVAL_ERROR)
ok($r==42) unless $dbh->$->{?mx_handle_list} && ok(1) # skip for Multiplex

$dbh->setHandleError(undef)
ok(!$dbh->HandleError)

do
    # dump_results;
    my $sth = $dbh->prepare($std_sql)

    isa_ok($sth, "DBI::st")

    if ($haveFileSpec && length(File::Spec->updir))
        ok($sth->execute(File::Spec->updir))
    else
        ok($sth->execute('../'))
    

    my $dump_dir = (env::var('TMP')
                 || env::var('TEMP')
                 || env::var('TMPDIR')
                 || env::var('SYS$SCRATCH')
                 || '/tmp')
    my $dump_file = ($haveFileSpec) ??
        File::Spec->catfile($dump_dir, 'dumpcsr.tst')
        !!
        "$dump_dir/dumpcsr.tst"
    #($dump_file) = ($dump_file =~ m/^(.*)$/);  # untaint

    SKIP: do
        skip "# dump_results test skipped: unable to open $dump_file: $^OS_ERROR\n", 2 unless (open(my $dump_results, ">", "$dump_file"))
        ok($sth->dump_results("10", "\n", ",\t", \$dump_results->*))
        close($dump_results)
        ok(-s $dump_file +> 0)
    

    is( unlink( $dump_file ), 1, "Remove $dump_file" )
    ok( !-e $dump_file, "Actually gone" )



diag "table_info\n"
# First generate a list of all subdirectories
$dir = getcwd()
ok(opendir(my $dh, $dir))
my(%dirs, %unexpected, %missing)
while (defined(my $file = readdir($dh)))
    %dirs{+$file} = 1 if -d $file

diag "Local $dir subdirs: $(join ' ', keys %dirs )\n"
closedir($dh)
my $sth = $dbh->table_info($dir, undef, "\%", "TABLE")
ok($sth)
%unexpected = %:  < %dirs 
%missing = %:  () 
while (my $ref = $sth->fetchrow_hashref())
    if (exists(%unexpected{$ref->{?'TABLE_NAME'}}))
        delete %unexpected{$ref->{?'TABLE_NAME'}}
    else
        %missing{+$ref->{?'TABLE_NAME'}} = 1
    

ok( ! %unexpected )
    or diag "Unexpected directories: ", join(",", keys %unexpected), "\n"
ok( ! %missing )
    or diag "Missing directories: ", join(",", keys %missing), "\n"

diag "tables\n"
my @tables_expected = @: 
    q{"schema"."table"}
    q{"sch-ema"."table"}
    q{"schema"."ta-ble"}
    q{"sch ema"."table"}
    q{"schema"."ta ble"}
    
my @tables = $dbh->tables(undef, undef, "\%", "VIEW")
ok(nelems(@tables) == nelems(@tables_expected), "Table count mismatch".nelems(@tables_expected)." vs ".nelems(@tables))
foreach (0..(nelems(@tables_expected)-1))
    ok(@tables[$_] eq @tables_expected[$_], "@tables[$_] ne @tables_expected[$_]")

my $i = 0
while ( $i +< 300 )
    diag "Testing the fake directories ($i).\n"
    ok($csr_a = $dbh->prepare("SELECT name, mode FROM long_list_$i"))
    ok($csr_a->execute(), DBI::errstr)
    my $ary = $csr_a->fetchall_arrayref
    ok(nelems($ary->@) == $i, nelems($ary->@)." rows instead of $i")
    if ($i)
        my @n1 = map { $_->[0] }, $ary->@
        my @n2 = reverse map { "file$_" }, 1..$i
        is("$(join ' ', @n1)", "$(join ' ', @n2)")
    else
        ok(1)
    
    $i += 100



SKIP: do
    skip "test not tested with Multiplex", 1
        if $dbh->$->{?mx_handle_list}
    diag "Testing \$dbh->func().\n"
    my %tables
    %tables = %+: map { $_ =~ m/lib/ ?? (%: $_, 1) !! (%: ) }, $dbh->tables()
    my @func_tables = $dbh->func('lib', 'examplep_tables')
    foreach my $t ( @func_tables)
        defined(delete %tables{$t}) or diag "Unexpected table: $t\n"
    
    is_deeply(%tables, (%: ))


$dbh->disconnect
ok(!$dbh->Active)
ok(!$dbh->ping, "ping should return false after disconnect")

1
