#!perl -w                                         # -*- perl -*-
# vim:sw=4:ts=8

use DBI
use Config
require VMS::Filespec if $^OS_NAME eq 'VMS'
require Cwd
use Test::More

my $haveFileSpec = try { require File::Spec }

$^OUTPUT_AUTOFLUSH = 1
$^WARNING = 1

# $\ = "\n"; # XXX Triggers bug, check this later (JW, 1998-12-28)

# Can we load the modules? If not, exit the test immediately:
# Reason is most probable a missing prerequisite.
#
# Is syslog available (required for the server)?

try {
    local $^WARN_HOOK = sub (@< @_) { $^EVAL_ERROR = shift };
    require DBD::Proxy;
    require DBI::ProxyServer;
    require Net::Daemon::Test;
}
if ($^EVAL_ERROR)
    if ($^EVAL_ERROR->{?description} =~ m/^Can't locate (\S+)/)
        plan skip_all => "modules required for proxy are probably not installed (e.g., $1)"
    
    die $^EVAL_ERROR


if ($DBI::PurePerl)
    # XXX temporary I hope
    plan skip_all => "DBD::Proxy currently has a problem under DBI::PurePerl"


# Create an empty config file to make sure that settings aren't
# overloaded by /etc/dbiproxy.conf
my $config_file = "dbiproxytst.conf"
unlink $config_file
my $fh
(open($fh, ">$config_file")  and
 (print $fh, "\{\}\n")          and
 close($fh))
    or die "Failed to create config file $config_file: $^OS_ERROR"

my($handle, $port)
my $numTests = 131

if (nelems(@ARGV))
    $port = @ARGV[0]
else

    # set DBI_TRACE to 0 to just get dbiproxy.log DBI trace for server
    # set DBI_TRACE > 0 to also get DBD::Proxy trace
    unlink "dbiproxy.log"
    unlink "dbiproxy.truss"

    # Uncommentand adjust this to isolate pure-perl client from server settings:
    # local %ENV{DBI_PUREPERL} = 0;

    # If desperate uncomment this and add '-d' after $^X below:
    # local %ENV{PERLDB_OPTS} = "AutoTrace NonStop=1 LineInfo=dbiproxy.dbg";

    # pass our @INC to children (e.g., so -Mblib passes through)
    env::var('PERL5LIB' ) = join(config_value("path_sep"), $^INCLUDE_PATH)

    my $dbi_trace_level = DBI->trace(0)
    my @child_args = @: 
        #'truss', '-o', 'dbiproxy.truss',
        $^EXECUTABLE_NAME, 'dbiproxy', '--test' # --test must be first command line arg
        ($dbi_trace_level ?? ('--dbitrace=dbiproxy.log') !! ())
        '--configfile', $config_file
        (($dbi_trace_level) ?? ('--logfile=1') !! ())
        '--mode=single'
        '--debug'
        '--timeout=60'
        
    warn " starting test dbiproxy process: $(join ' ', @child_args)\n" if $dbi_trace_level
    (@: $handle, $port) =  Net::Daemon::Test->Child($numTests, < @child_args)


my $debug = env::var('DBI_TRACE') ?? 1 !! 0
my $dsn = "DBI:Proxy:hostname=127.0.0.1;port=$port;debug=$debug;dsn=DBI:ExampleP:"

diag "Making a first connection and closing it immediately.\n"
ok(try { DBI->connect($dsn, '', '', \(%:  'PrintError' => 1 )) })
    or diag "Connect error: " . DBI::errstr . "\n"

diag "Making a second connection.\n"
my $dbh
ok($dbh = try { DBI->connect($dsn, '', '', \(%:  'PrintError' => 0 )) })
    or diag "Connect error: " . DBI::errstr . "\n"

diag "example_driver_path=$dbh->{?example_driver_path}\n"
ok($dbh->{?example_driver_path})

diag "Setting AutoCommit\n"
$^EVAL_ERROR = "old-error"	# should be preserved across DBI calls
ok($dbh->{+AutoCommit} = 1)
ok($dbh->{?AutoCommit})
ok($^EVAL_ERROR eq "old-error", "\$^EVAL_ERROR now '$^EVAL_ERROR'")
#$dbh->trace(2);

try {
    local $dbh->{+AutoCommit } = 1;   # This breaks die!
    die "BANG!!!\n";
}
ok($^EVAL_ERROR eq "BANG!!!\n", "\$^EVAL_ERROR value lost")


diag "begin_work...\n"
ok($dbh->{?AutoCommit})
ok(!$dbh->{?BegunWork})

ok($dbh->begin_work)
ok(!$dbh->{?AutoCommit})
ok($dbh->{?BegunWork})

$dbh->commit
ok(!$dbh->{?BegunWork})
ok($dbh->{?AutoCommit})

ok($dbh->begin_work(\(%: )))
$dbh->rollback
ok($dbh->{?AutoCommit})
ok(!$dbh->{?BegunWork})


diag "Doing a ping.\n"
$_ = $dbh->ping
ok($_)
ok($_ eq '2') # ping was DBD::ExampleP's ping

diag "Ensure CompatMode enabled.\n"
ok($dbh->{?CompatMode})

diag "Trying local quote.\n"
$dbh->{+'proxy_quote'} = 'local'
ok($dbh->quote("quote's") eq "'quote''s'")
ok($dbh->quote(undef)     eq "NULL")

diag "Trying remote quote.\n"
$dbh->{+'proxy_quote'} = 'remote'
ok($dbh->quote("quote's") eq "'quote''s'")
ok($dbh->quote(undef)     eq "NULL")

# XXX the $optional param is undocumented and may be removed soon
ok($dbh->quote_identifier('foo')    eq '"foo"',  $dbh->quote_identifier('foo'))
ok($dbh->quote_identifier('f"o')    eq '"f""o"', $dbh->quote_identifier('f"o'))
ok($dbh->quote_identifier('foo','bar') eq '"foo"."bar"')
ok($dbh->quote_identifier('foo',undef,'bar') eq '"foo"."bar"')
ok($dbh->quote_identifier(undef,undef,'bar') eq '"bar"')

diag "Trying commit with invalid number of parameters.\n"
try { $dbh->commit('dummy') }
ok($^EVAL_ERROR =~ m/^DBI commit: invalid number of arguments:/)
    unless $DBI::PurePerl && ok(1)

diag "Trying select with unknown field name.\n"
my $cursor_e = $dbh->prepare("select unknown_field_name from ?")
ok(defined $cursor_e)
ok(!$cursor_e->execute('a'))
ok(DBI::err)
ok(DBI::err == $dbh->err)
ok(DBI::errstr =~ m/unknown_field_name/, DBI::errstr)

ok(DBI::errstr eq $dbh->errstr)
ok($dbh->errstr eq $dbh->func('errstr'))

my $dir = Cwd::cwd()	# a dir always readable on all platforms
$dir = VMS::Filespec::unixify($dir) if $^OS_NAME eq 'VMS'

diag "Trying a real select.\n"
my $csr_a = $dbh->prepare("select mode,size,name from ?")
ok(ref $csr_a)
ok($csr_a->execute($dir))
    or diag "Execute failed: ", < $csr_a->errstr(), "\n"

diag "Repeating the select with second handle.\n"
my $csr_b = $dbh->prepare("select mode,size,name from ?")
ok(ref $csr_b)
ok($csr_b->execute($dir))
ok($csr_a != $csr_b)
ok($csr_a->{?NUM_OF_FIELDS} == 3)
if ($DBI::PurePerl)
    $csr_a->trace(2)
    use Data::Dumper;
    warn < Dumper($csr_a->{?Database})

ok($csr_a->{Database}->{Driver}->{?Name} eq 'Proxy', "Name=$csr_a->{Database}->{Driver}->{?Name}")
$csr_a->trace(0), die if $DBI::PurePerl

my($col0, $col1, $col2)
my(@row_a, @row_b)

#$csr_a->trace(2);
diag "Trying bind_columns.\n"
ok($csr_a->bind_columns(undef, \$col0, \$col1, \$col2) )
ok($csr_a->execute($dir))
@row_a = $csr_a->fetchrow_array
ok(nelems(@row_a))
ok(@row_a[0] eq $col0)
ok(@row_a[1] eq $col1)
ok(@row_a[2] eq $col2)

diag "Trying bind_param.\n"
ok($csr_b->bind_param(1, $dir))
ok($csr_b->execute())
@row_b =  $csr_b->fetchrow_arrayref->@
ok(nelems(@row_b))

ok("$(join ' ', @row_a)" eq "$(join ' ', @row_b)")
@row_b = $csr_b->fetchrow_array
ok("$(join ' ', @row_a)" ne "$(join ' ', @row_b)")
    or diag("Expected something different from '\%s', got '\%s'\n", "$(join ' ', @row_a)",
            "$(join ' ', @row_b)")

diag "Trying fetchrow_hashref.\n"
ok($csr_b->execute())
my $row_b = $csr_b->fetchrow_hashref
ok($row_b)
diag "row_a: $(join ' ', @row_a)\n"
diag "row_b: $(join ' ', $row_b->% )\n"
ok($row_b->{?mode} == @row_a[0])
ok($row_b->{?size} == @row_a[1])
ok($row_b->{?name} eq @row_a[2])

diag "Trying fetchrow_hashref with FetchHashKeyName.\n"
do
    #local $dbh->{TraceLevel} = 9;
    local $dbh->{+FetchHashKeyName} = 'NAME_uc'
    ok($dbh->{?FetchHashKeyName} eq 'NAME_uc')
    my $csr_c = $dbh->prepare("select mode,size,name from ?")
    ok($csr_c->execute($dir), DBI::errstr)
    $row_b = $csr_c->fetchrow_hashref
    ok($row_b)
    diag "row_b: $(join ' ', $row_b->% )\n"
    ok($row_b->{?MODE} eq @row_a[0])


diag "Trying finish.\n"
ok($csr_a->finish)
#ok($csr_b->finish);
ok(1)

diag "Forcing destructor.\n"
$csr_a = undef  # force destruction of this cursor now
ok(1)

diag "Trying fetchall_arrayref.\n"
ok($csr_b->execute())
my $r = $csr_b->fetchall_arrayref
ok($r)
ok(nelems($r->@))
ok($r->[0]->[0] == @row_a[0])
ok($r->[0]->[1] == @row_a[1])
ok($r->[0]->[2] eq @row_a[2])

ok($csr_b->finish)


diag "Retrying unknown field name.\n"
my $csr_c
$csr_c = $dbh->prepare("select unknown_field_name1 from ?")
ok($csr_c)
ok(!$csr_c->execute($dir))
ok(DBI::errstr =~ m/Unknown field names: unknown_field_name1/)
    or diag("Wrong error string: \%s", DBI::errstr)

diag "Trying RaiseError.\n"
$dbh->{+RaiseError} = 1
ok($dbh->{?RaiseError})
ok($csr_c = $dbh->prepare("select unknown_field_name2 from ?"))
ok(!try { $csr_c->execute(); 1 })
#diag "$^EVAL_ERROR\n";
ok($^EVAL_ERROR =~ m/Unknown field names: unknown_field_name2/)
$dbh->{+RaiseError} = 0
ok(!$dbh->{?RaiseError})

diag "Trying warnings.\n"
do
    my @warn
    local $^WARN_HOOK = sub (@< @_) { push @warn, < @_ }
    $dbh->{+PrintError} = 1
    ok($dbh->{?PrintError})
    ok(($csr_c = $dbh->prepare("select unknown_field_name3 from ?")))
    ok(!$csr_c->execute())
    ok("$(join ' ', @warn)" =~ m/Unknown field names: unknown_field_name3/)
    $dbh->{+PrintError} = 0
    ok(!$dbh->{?PrintError})

$csr_c->finish()


diag "Trying type_info_all.\n"
my $array = $dbh->type_info_all()
ok($array  and  ref($array) eq 'ARRAY')
    or diag("Expected ARRAY, got \%s, error \%s\n", < DBI::neat($array), <
            $dbh->errstr())
ok($array->[0]  and  ref($array->[0]) eq 'HASH')
my $ok = 1
for my $i (1..nelems($array->@)-1)
    diag "$array->[$i]\n"
    $ok = 0  unless ($array->[$i]  and  ref($array->[$i]) eq 'ARRAY')
    diag "$ok\n"

ok($ok)

# Test the table_info method
# First generate a list of all subdirectories
$dir = $haveFileSpec ?? File::Spec->curdir() !! "."
ok(opendir(my $dh, $dir))
my(%dirs, %unexpected, %missing)
while (defined(my $file = readdir($dh)))
    %dirs{+$file} = 1 if -d $file

closedir($dh)
my $sth = $dbh->table_info(undef, undef, undef, undef)
ok($sth) or warn "table_info failed: ", < $dbh->errstr(), "\n"
%missing = %:  < %dirs 
%unexpected = %:  () 
while (my $ref = $sth->fetchrow_hashref())
    diag "table_info: Found table $ref->{?'TABLE_NAME'}\n"
    if (exists(%missing{$ref->{?'TABLE_NAME'}}))
        delete %missing{$ref->{?'TABLE_NAME'}}
    else
        %unexpected{+$ref->{?'TABLE_NAME'}} = 1
    

ok(!$sth->errstr())
    or diag "Fetching table_info rows failed: ", < $sth->errstr(), "\n"
ok(keys %unexpected == 0)
    or diag "Unexpected directories: ", join(",", keys %unexpected), "\n"
ok(keys %missing == 0)
    or diag "Missing directories: ", join(",", keys %missing), "\n"

# Test the tables method
%missing = %:  < %dirs 
%unexpected = %:  () 
diag "Expecting directories ", join(",", keys %dirs), "\n"
foreach my $table ( $dbh->tables())
    diag "tables: Found table $table\n"
    if (exists(%missing{$table}))
        delete %missing{$table}
    else
        %unexpected{+$table} = 1
    

ok(!$sth->errstr())
    or diag "Fetching table_info rows failed: ", < $sth->errstr(), "\n"
ok(keys %unexpected == 0)
    or diag "Unexpected directories: ", join(",", keys %unexpected), "\n"
ok(keys %missing == 0)
    or diag "Missing directories: ", join(",", keys %missing), "\n"


# Test large recordsets
my $i = 0
while ($i +< 300)
    diag "Testing the fake directories ($i).\n"
    ok($csr_a = $dbh->prepare("SELECT name, mode FROM long_list_$i"))
    ok($csr_a->execute(), DBI::errstr)
    my $ary = $csr_a->fetchall_arrayref
    ok(nelems($ary->@) == $i)
    if ($i)
        my @n1 = map { $_->[0] }, $ary->@
        my @n2 = reverse map { "file$_" }, 1..$i
        ok("$(join ' ', @n1)" eq "$(join ' ', @n2)")
    else
        ok(1)
    
    $i += 100



# Test the RowCacheSize attribute
ok($csr_a = $dbh->prepare("SELECT * FROM ?"))
ok($dbh->{?'RowCacheSize'} == 20)
ok($csr_a->{?'RowCacheSize'} == 20)
ok($csr_a->execute('long_list_50'))
ok($csr_a->fetchrow_arrayref())
ok($csr_a->{?'proxy_data'}  and  nelems($csr_a->{?'proxy_data'}->@) == 19)
ok($csr_a->finish())

ok($dbh->{+'RowCacheSize'} = 30)
ok($dbh->{?'RowCacheSize'} == 30)
ok($csr_a->{?'RowCacheSize'} == 30)
ok($csr_a->execute('long_list_50'))
ok($csr_a->fetchrow_arrayref())
ok($csr_a->{?'proxy_data'}  and  nelems($csr_a->{?'proxy_data'}->@) == 29)
    or diag("Expected 29 records in cache, got " . nelems($csr_a->{?'proxy_data'}->@) .
            "\n")
ok($csr_a->finish())


ok($csr_a->{+'RowCacheSize'} = 10)
ok($dbh->{?'RowCacheSize'} == 30)
ok($csr_a->{?'RowCacheSize'} == 10)
ok($csr_a->execute('long_list_50'))
ok($csr_a->fetchrow_arrayref())
ok($csr_a->{?'proxy_data'}  and  nelems($csr_a->{?'proxy_data'}->@) == 9)
    or diag("Expected 9 records in cache, got " . nelems($csr_a->{?'proxy_data'}->@) .
            "\n")
ok($csr_a->finish())

$dbh->disconnect

# Test $dbh->func()
#  diag "Testing \$dbh->func().\n";
#  my %tables = map { $_ =~ /lib/ ? ($_, 1) : () } $dbh->tables();
#  $ok = 1;
#  foreach my $t ($dbh->func('lib', 'examplep_tables')) {
#      defined(delete %tables{$t}) or diag "Unexpected table: $t\n";
#  }
#  ok(%tables == 0);



END 
    my $status = $^CHILD_ERROR
    $handle->Terminate() if $handle
    undef $handle
    my $f = $config_file
    undef $config_file
    unlink $f if $f
    $^CHILD_ERROR = $status
;

1
