#!perl -w                                         # -*- perl -*-
# vim:sw=4:ts=8
$^OUTPUT_AUTOFLUSH=1

use warnings

use DBI
use Data::Dumper
use Test::More

# here we test the DBI_GOFER_RANDOM mechanism
# and how gofer deals with failures

plan skip_all => "requires Callbacks which are not supported with PurePerl" if $DBI::PurePerl

if (my $ap = env::var('DBI_AUTOPROXY')) # limit the insanity
    plan skip_all => "Gofer DBI_AUTOPROXY" if $ap =~ m/^dbi:Gofer/i

    # this means we have DBD::Gofer => DBD::Gofer => DBD::whatever
    # rather than disable it we let it run because we're twisted
    # and because it helps find more bugs (though debugging can be painful)
    warn "\n$^PROGRAM_NAME is running with DBI_AUTOPROXY enabled ($(env::var('DBI_AUTOPROXY')))\n"
        unless $^PROGRAM_NAME =~ m/\bzv/ # don't warn for t/zvg_85gofer.t


plan 'no_plan'

TODO: do
    todo_skip "Failure for unknown reasons starting after Perl Kurila 1.12"

    my $tmp
    my $dbh
    my $fails

    # we'll use the null transport for simplicity and speed
    # and the rush policy to limit the number of interactions with the gofer executor

    # silence the "DBI_GOFER_RANDOM..." warnings
    my @warns
    $^WARN_HOOK = sub (@< @_)
        my $msg = @_[0]->message
        ($msg =~ m/^DBI_GOFER_RANDOM/) ?? push(@warns, $msg) !! print $^STDERR, $msg
    

    # --- 100% failure rate

    (@: $fails, $dbh) =  trial_impact("fail=100%,do", 10, "", sub (@< @_) { $_->do("set foo=1") })
    is $fails, 100, 'should fail 100% of the time'
    ok   $^EVAL_ERROR, '$^EVAL_ERROR should be set'
    like $^EVAL_ERROR->{?description}, '/fake error induced by DBI_GOFER_RANDOM/'
    ok   $dbh->errstr, 'errstr should be set'
    like $dbh->errstr, '/DBI_GOFER_RANDOM/', 'errstr should contain DBI_GOFER_RANDOM'
    ok !$dbh->{go_response}->executed_flag_set, 'go_response executed flag should be false'


    # XXX randomness can't be predicted, so it's just possible these will fail
    srand(42) # try to limit occasional failures (effect will vary by platform etc)

    sub trial_impact($spec, $count, $dsn_attr, $code)
        local env::var('DBI_GOFER_RANDOM' ) = $spec
        my $dbh = dbi_connect("policy=rush;$dsn_attr")
        local $_ = $dbh
        my $fail_percent = percentage_exceptions(200, $code)
        return @: $fail_percent, $dbh 
    

    # --- 50% failure rate, with no retries

    $fails = trial_impact("fail=50%,do", 200, "retry_limit=0", sub (@< @_) { $_->do("set foo=1") })[0]
    diag "target approx 50% random failures, got $fails%\n"
    between_ok($fails, 10, 90, "should fail about 50% of the time, but at least between 10% and 90%")

    # --- 50% failure rate, with many retries (should yield low failure rate)

    $fails = trial_impact("fail=50%,prepare", 200, "retry_limit=5", sub (@< @_) { $_->prepare("set foo=1") })[0]
    diag "target less than 20% effective random failures (ideally 0), got $fails%\n"
    cmp_ok $fails, '+<', 20, 'should fail < 20%'

    # --- 10% failure rate, with many retries (should yield zero failure rate)

    $fails = trial_impact("fail=10,do", 200, "retry_limit=10", sub (@< @_) { $_->do("set foo=1") })[0]
    cmp_ok $fails, '+<', 1, 'should fail < 1%'

    # --- 50% failure rate, test is_idempotent

    env::var("DBI_GOFER_RANDOM") = "fail=50%,do"   # 50%

    # test go_retry_hook and that ReadOnly => 1 retries a non-idempotent statement
    ok(my $dbh_50r1ro = dbi_connect("policy=rush;retry_limit=1", \(%: 
                                go_retry_hook => sub (@< @_) { return (@_[0]->is_idempotent) ?? 1 !! 0 }
                                ReadOnly => 1
                                ) ))
    between_ok(< percentage_exceptions(100, sub (@< @_) { $dbh_50r1ro->do("set foo=1") }),
           10, 40, 'should fail ~25% (ie 50% with one retry)')
    between_ok($dbh_50r1ro->{go_transport}->meta->{?request_retry_count},
           20, 80, 'transport request_retry_count should be around 50')

    # test as above but with ReadOnly => 0
    ok(my $dbh_50r1rw = dbi_connect("policy=rush;retry_limit=1", \(%: 
                                go_retry_hook => sub (@< @_) { return (@_[0]->is_idempotent) ?? 1 !! 0 }
                                ReadOnly => 0
                                ) ))
    between_ok(< percentage_exceptions(100, sub (@< @_) { $dbh_50r1rw->do("set foo=1") }),
           20, 80, 'should fail ~50%, ie no retries')
    ok !$dbh_50r1rw->{go_transport}->meta->{?request_retry_count},
        'transport request_retry_count should be zero or undef'


    # --- check random is random and non-random is non-random

    my %fail_percents
    for (1..5)
        $fails = trial_impact("fail=50%,do", 10, "", sub (@< @_) { $_->do("set foo=1") })[0]
        ++%fail_percents{+$fails}
    
    cmp_ok scalar keys %fail_percents, '+>=', 2, 'positive percentage should fail randomly'

    %fail_percents = %:  () 
    for (1..5)
        $fails = trial_impact("fail=-50%,do", 10, "", sub (@< @_) { $_->do("set foo=1") })[0]
        ++%fail_percents{+$fails}
    
    is scalar keys %fail_percents, 1, 'negative percentage should fail non-randomly'

    # ---
    diag "Testing random delay\n"

    env::var("DBI_GOFER_RANDOM") = "delay0.1=51%,do" # odd percentage to force warn()s
    @warns = $@
    ok($dbh = dbi_connect("policy=rush;retry_limit=0"))
    is percentage_exceptions(20, sub (@< @_) { $dbh->do("set foo=1") }),
        0, "should not fail for DBI_GOFER_RANDOM='$(env::var('DBI_GOFER_RANDOM'))'"
    my $delays = grep { m/delaying execution/ }, @warns
    between_ok($delays, 1, 19, 'should be delayed around 5 times')

    exit 0

    # --- subs ---
    #
    sub between_ok($got, $min, $max, $label)
        local $Test::Builder::Level = 2
        cmp_ok $got, '+>=', $min, "$label (got $got)"
        cmp_ok $got, '+<=', $max, "$label (got $got)"
    

    sub dbi_connect($gdsn, $attr)
        return DBI->connect("dbi:Gofer:transport=null;$gdsn;dsn=dbi:ExampleP:", 0, 0, \(%: 
            RaiseError => 1, PrintError => 0, ($attr) ?? < $attr->% !! ()
            ))
    

    sub percentage_exceptions($count, $sub)
        my $i = $count
        my $exceptions = 0
        while ($i--)
            try { $sub->() }
            if ($^EVAL_ERROR)
                die "Unexpected failure: $($^EVAL_ERROR->message)" unless $^EVAL_ERROR->{?description} =~ m/DBI_GOFER_RANDOM/
                ++$exceptions
            
        
        return $exceptions/$count*100
    


