#!/usr/bin/perl -w

use env
use DateTime::TimeZone::Local
use DateTime::TimeZone::Local::Unix
use File::Basename < qw( basename )
use File::Spec
use Sys::Hostname
use Test::More

use lib File::Spec->catdir( File::Spec->curdir, 't' )

BEGIN { require 'check_datetime_version.pl' }

my @aliases = sort  keys DateTime::TimeZone::links()
my @names = DateTime::TimeZone::all_names

plan tests => nelems(@aliases) + nelems(@names) + 22


do
    for my $alias ( sort  @aliases )
        local env::var('TZ' ) = $alias
        my $tz = DateTime::TimeZone::Local->TimeZone()
        isa_ok( $tz, 'DateTime::TimeZone' )
    


do
    for my $name (sort @names)
        local env::var('TZ') = $name
        my $tz = try { DateTime::TimeZone::Local->TimeZone() }
        isa_ok( $tz, 'DateTime::TimeZone' )
    


do
    local env::var('TZ') = 'this will not work'

    my $tz
    try { $tz = DateTime::TimeZone::Local::Unix->FromEnv() }
    is( $tz, undef,
        'invalid time zone name in $ENV{TZ} fails' )

    local env::var('TZ' ) = '123/456'

    try { $tz = DateTime::TimeZone::Local::Unix->FromEnv() }
    is( $tz, undef,
        'invalid time zone name in $ENV{TZ} fails' )


do
    local env::var('TZ') = 'Africa/Kinshasa'

    my $tz
    try { $tz = DateTime::TimeZone::Local::Unix->FromEnv() }
    is( $tz->name(), 'Africa/Kinshasa', 'tz object name() is Africa::Kinshasa' )


SKIP:
    do
    skip "/etc/localtime is not a symlink", 2
        unless -l '/etc/localtime'

    $^WARNING = 0
    local *DateTime::TimeZone::Local::Unix::_Readlink = sub (@< @_) { '/usr/share/zoneinfo/US/Eastern' }
    $^WARNING = 1

    my $tz
    try { $tz = DateTime::TimeZone::Local::Unix->FromEtcLocaltime() }
    is( $^EVAL_ERROR, '', 'valid time zone name in /etc/localtime symlink should not die' )
    isa_ok( $tz, 'DateTime::TimeZone::America::New_York' )


SKIP:
    do
    skip "cannot read /etc/sysconfig/clock", 2
        unless -r '/etc/sysconfig/clock' && -f _

    $^WARNING = 0
    local *DateTime::TimeZone::Local::Unix::_ReadEtcSysconfigClock = sub (@< @_) { 'US/Eastern' }
    $^WARNING = 1

    my $tz
    try { $tz = DateTime::TimeZone::Local::Unix->FromEtcSysconfigClock() }
    is( $^EVAL_ERROR, '', 'valid time zone name in /etc/sysconfig/clock should not die' )
    isa_ok( $tz, 'DateTime::TimeZone::America::New_York' )


SKIP:
    do
    skip "cannot read /etc/default/init", 2
        unless -r '/etc/default/init' && -f _

    local *DateTime::TimeZone::Local::_read_etc_default_init = sub (@< @_) { 'US/Eastern' }

    my $tz
    try { $tz = DateTime::TimeZone::Local::Unix->FromEtcDefaultInit() }
    is( $^EVAL_ERROR, '', 'valid time zone name in /etc/default/init should not die' )
    isa_ok( $tz, 'DateTime::TimeZone::Australia::Melbourne' )


SKIP:
    do
    skip "Cannot run these tests without explicitly knowing local time zone first (only runs on developers' machine)", 6
        unless hostname =~ m/houseabsolute/ && -d '.svn'

    do
        local env::var('TZ') = ''

        my $tz
        try { $tz = DateTime::TimeZone::Local->TimeZone() }
        is( $^EVAL_ERROR, '', 'valid time zone name in /etc/localtime should not die' )
        isa_ok( $tz, 'DateTime::TimeZone::America::Chicago' )
    

    do
        $^WARNING = 0
        local *DateTime::TimeZone::Local::Unix::FromEtcLocaltime = sub (@< @_) { undef }
        $^WARNING = 1

        my $tz
        try { $tz = DateTime::TimeZone::Local->TimeZone() }
        is( $^EVAL_ERROR, '', 'valid time zone name in /etc/timezone should not die' )
        isa_ok( $tz, 'DateTime::TimeZone::America::Chicago' )
    

    do
        # requires that /etc/default/init contain
        # TZ=Australia/Melbourne to work.
        $^WARNING = 0
        local *DateTime::TimeZone::Local::Unix::FromEtcLocaltime = sub (@< @_) { undef }
        local *DateTime::TimeZone::Local::Unix::FromEtcTimezone = sub (@< @_) { undef }
        local *DateTime::TimeZone::Local::Unix::FromEtcTIMEZONE = sub (@< @_) { undef }
        $^WARNING = 1

        my $tz
        try { $tz = DateTime::TimeZone::Local->TimeZone() }
        is( $^EVAL_ERROR, '', '/etc/default/init contains TZ=Australia/Melbourne' )
        isa_ok( $tz, 'DateTime::TimeZone::Australia::Melbourne' )
    


SKIP:
    do
    skip "These tests are too dangerous to run on someone else's machine ;)", 4
        unless hostname =~ m/houseabsolute|quasar/ && -d '.svn'

    skip "These tests can only be run if we can overwrite /etc/localtime", 4
        unless -w '/etc/localtime' && -l '/etc/localtime'

    my $tz_file = readlink '/etc/localtime'

    unlink '/etc/localtime' or die "Cannot unlink /etc/localtime: $^OS_ERROR"

    require File::Copy
    File::Copy::copy( '/usr/share/zoneinfo/Asia/Calcutta', '/etc/localtime' )
        or die "Cannot copy /usr/share/zoneinfo/Asia/Calcutta to '/etc/localtime': $^OS_ERROR"

    do
        local env::var('TZ') = ''

        require Cwd
        my $cwd = Cwd::cwd()

        my $tz
        try { $tz = DateTime::TimeZone::Local->TimeZone() }
        is( $^EVAL_ERROR, '', 'copy of zoneinfo file at /etc/localtime' )
        isa_ok( $tz, 'DateTime::TimeZone::Asia::Calcutta' )

        is( Cwd::cwd(), $cwd, 'cwd should not change after finding local time zone' )
    

    do
        local env::var('TZ') = ''

        # Make sure that a die handler does not break our use of die
        # to escape from File::Find::find()
        local $^DIE_HOOK = sub (@< @_) { die 'haha'; }

        my $tz
        try { $tz = DateTime::TimeZone::Local->TimeZone() }
        isa_ok( $tz, 'DateTime::TimeZone::Asia::Calcutta' )
    

    unlink '/etc/localtime' or die "Cannot unlink /etc/localtime: $^OS_ERROR"
    symlink $tz_file, '/etc/localtime'
        or die "Cannot symlink $tz_file to '/etc/localtime': $^OS_ERROR"


do
    local env::var('TZ') = 'Australia/Melbourne'
    my $tz = try { DateTime::TimeZone->new( name => 'local' ) }
    isa_ok( $tz, 'DateTime::TimeZone::Australia::Melbourne' )


SKIP:
    do
    skip "These tests only run on Win32", 1
        unless $^OS_NAME =~ m/win32/i

    require DateTime::TimeZone::Local::Win32

    my %Reg
    Win32::TieRegistry->import( TiedHash => \%Reg )

    local %Reg{+'HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\TimeZoneInformation\StandardName'}
        = 'Eastern Standard Time'

    my $tz
    try { $tz = DateTime::TimeZone::Local::Win32->FromRegistry() }
    isa_ok( $tz, 'DateTime::TimeZone::America::New_York' )



SKIP:
    do
    skip "These tests require File::Temp", 1
        unless require File::Temp
    skip "These tests require a filesystem which support symlinks", 1
        unless try { symlink '', '' ; 1 }

    my $tempdir = File::Temp::tempdir( CLEANUP => 1 )

    my $first = File::Spec->catfile( $tempdir, 'first' )
    open my $fh, '>', $first
        or die "Cannot open $first: $^OS_ERROR"
    close $fh

    my $second = File::Spec->catfile( $tempdir, 'second' )
    symlink $first => $second
        or die "Cannot symlink $first => $second: $^OS_ERROR"

    my $third = File::Spec->catfile( $tempdir, 'third' )
    symlink $second => $third
        or die "Cannot symlink $first => $second: $^OS_ERROR"

    # It seems that on some systems (OSX, others?) the temp directory
    # returned by File::Temp may be a symlink (/tmp is a link to
    # /private/tmp), so when abs_path folows that link, we end up with
    # a different path to the "first" file.
    is( basename( DateTime::TimeZone::Local::Unix->_Readlink( $third ) ),
        basename( $first ),
        '_Readlink follows multiple levels of symlinking' )

