#!/usr/bin/perl -w

use Test::More tests => 42

use DateTime

my $pos = DateTime::Infinite::Future->new
my $neg = DateTime::Infinite::Past->new
my $posinf = 100 ** 1000
my $neginf = -1 * $posinf
# used to use abs() which broke some Win32 platforms but may have
# fixed others - will wait for bug reports
my $nan_string = ($posinf - $posinf) . ''

# infinite date math
do
    ok( $pos->is_infinite, 'positive infinity should be infinite' )
    ok( $neg->is_infinite, 'negative infinity should be infinite' )
    ok( !$pos->is_finite, 'positive infinity should not be finite' )
    ok( !$neg->is_finite, 'negative infinity should not be finite' )

    # that's a long time ago!
    my $long_ago = DateTime->new( year => -100_000 )

    is( $neg->compare($long_ago), -1,
        'negative infinity is really negative' )

    my $far_future = DateTime->new( year => 100_000 )
    is( $pos->compare($far_future), 1,
        'positive infinity is really positive' )

    is( $pos->compare($neg), 1,
        'positive infinity is bigger than negative infinity' )

    my $pos_dur = $pos->subtract_datetime($far_future)
    ok( $pos_dur->is_positive,
        'infinity - normal = infinity' )

    my $pos2 = $long_ago->clone()->add_duration($pos_dur)
    is( $pos2->compare($pos), 0,
        'normal + infinite duration = infinity' )

    my $neg_dur = $far_future->subtract_datetime($pos)
    ok( $neg_dur->is_negative,
        'normal - infinity = neg infinity' )

    my $neg2 = $long_ago->clone->add_duration($neg_dur)
    is( $neg2->compare($neg), 0,
        'normal + neg infinite duration = neg infinity' )

    my $dur = $pos->subtract_datetime($pos)
    my %deltas = %:  < $dur->deltas 
    my @compare = qw( days seconds nanoseconds )
    foreach ( @compare)
        # NaN != NaN (but should stringify the same)
        is( %deltas{?$_} . '', $nan_string,
            "infinity - infinity = nan ($_)" )
    

    my $new_pos = $pos->clone->add( days => 10 )
    is( $new_pos->compare($pos), 0,
        "infinity + normal duration = infinity" )

    my $new_pos2 = $pos->clone->subtract( days => 10 )
    is( $new_pos2->compare($pos), 0,
        "infinity - normal duration = infinity" )

    is( $pos->compare($posinf), 0,
        "infinity (datetime) == infinity (number)" )

    is( $neg->compare($neginf), 0,
        "neg infinity (datetime) == neg infinity (number)" )


# This could vary across platforms
my $pos_as_string = $posinf . ''
my $neg_as_string = $neginf . ''

# formatting
do
    foreach my $m ( qw( year month day hour minute second
                        microsecond millisecond nanosecond ) )
        is( $pos->?$m() . '', $pos_as_string,
            "pos $m is $pos_as_string" )

        is( $neg->?$m() . '', $neg_as_string,
            "neg $m is $pos_as_string" )
    


do
    my $now  = DateTime->now

    is( DateTime->compare($pos, $now),  1, 'positive infinite is greater than now' )
    is( DateTime->compare($neg, $now), -1, 'negative infinite is less than now' )


do
    my $now = DateTime->now
    my $pos2 = $pos->add_duration( DateTime::Duration->new( months => 1 ) )

    ok( $pos \== $pos2,
        "infinity (datetime) == infinity (datetime)" )


do
    my $now = DateTime->now
    my $neg2 = $neg->add_duration( DateTime::Duration->new( months => 1 ) )

    is( $neg->compare($neg2), 0,
        "-infinity (datetime) == -infinity (datetime)" )


do
    cmp_ok( $pos->_stringify, 'eq', $posinf, 'stringified infinity (datetime) eq infinity (number)' )
    cmp_ok( $neg->_stringify, 'eq', $neginf, 'stringified neg infinity (datetime) eq neg infinity (number)' )

