# Check that reassigning to a pending event wont remove it from the event queue. It
# should remain scheduled. But then also check that this 'dangling' event is
# properly cleaned up.

package MyEvent
use Event::Lib

use base < qw/Event::Lib::event/
use warnings

our @DESTROYS

sub new
    my $class = shift
    bless event_new(< @_) => $class


do
    my $count = 0
    sub DESTROY
        my $self = shift
        push @DESTROYS, ++$count
        no warnings 'misc';
        $self->SUPER::DESTROY
    


package MyTimer
use Event::Lib
use base < qw/Event::Lib::timer/

our @DESTROYS

sub new
    my $class = shift
    bless timer_new(< @_) => $class


do
    my $count = 0
    sub DESTROY
        my $self = shift
        push @DESTROYS, ++$count
        no warnings 'misc';
        $self->SUPER::DESTROY
    


package MySignal
use Event::Lib
use base < qw/Event::Lib::signal/

our @DESTROYS

sub new
    my $class = shift
    bless signal_new(< @_) => $class


do
    my $count = 0
    sub DESTROY
        my $self = shift
        push @DESTROYS, ++$count
        no warnings 'misc';
        $self->SUPER::DESTROY
    


package main

use IO::Socket::UNIX
use Event::Lib
use Test::More
use POSIX < qw/SIGHUP SIGUSR1/

BEGIN { plan tests => 5 + 4 + 4};

do   # fh-event
    my $PATH1 = "t/sock1-$^PID"
    my $PATH2 = "t/sock2-$^PID"
    if (fork)
        # PARENT

        my $sock1 = IO::Socket::UNIX->new(
            Local => $PATH1,
            Listen => 1,
            ) or die "$^EVAL_ERROR"
        $sock1->listen or die $^OS_ERROR
        my $sock2 = IO::Socket::UNIX->new(
            Local => $PATH2,
            Listen => 1,
            ) or die "$^EVAL_ERROR"
        $sock2->listen or die $^OS_ERROR

        open my $me, "<", $^PROGRAM_NAME or die $^OS_ERROR
        my $e = MyEvent->new($sock1, EV_READ, sub (@< @_) { ok(my $h = shift->fh->accept); $h->close; })
        $e->add
        ok($e->pending)
        $e = MyEvent->new($sock2, EV_READ, sub (@< @_) { ok(my $h = shift->fh->accept); $h->close; })
        $e->add

        is(@MyEvent::DESTROYS[0], 1)
        event_mainloop
        is(@MyEvent::DESTROYS[1], 2)

        for (@: $PATH1, $PATH2)
            unlink $_ 
        wait
    else
        # CHILD
        sleep 2
        my $real_client1 = IO::Socket::UNIX->new(Peer => $PATH1) or die $^EVAL_ERROR
        sleep 1
        my $real_client2 = IO::Socket::UNIX->new(Peer => $PATH2) or die $^EVAL_ERROR
        for (@: $real_client1, $real_client2)
            $_->close 
        exit


do   # timer-event
    my $e = MyTimer->new(sub (@< @_) { ok(1); })
    $e->add(1)
    $e = MyTimer->new(sub (@< @_) { ok(1); })
    $e->add(2)

    is(@MyTimer::DESTROYS[0], 1)
    event_mainloop
    is(@MyTimer::DESTROYS[1], 2)


do   # signal-event
    my $parentpid = $^PID
    if (my $pid = fork)
        my $e = MySignal->new(SIGHUP, sub (@< @_) { ok(1, "SIGHUP signal received"); shift->remove; })
        $e->add
        $e = MySignal->new(SIGUSR1, sub (@< @_) { ok(1, "SIGUSR1 signal received"); shift->remove; })
        $e->add

        is(@MyTimer::DESTROYS[0], 1, "Destroy is 1")
        event_mainloop
        is(@MyTimer::DESTROYS[1], 2)
        wait
    else
        # CHILD
        sleep 1
        kill SIGHUP     => $parentpid
        kill SIGUSR1    => $parentpid
        exit
    

