#!/usr/bin/perl -w
use warnings

use Test::More

use MIME::Tools
use File::Path
use File::Spec
use File::Basename
use MIME::WordDecoder < qw(unmime)

use lib < qw( ./t/ )
use Globby

use MIME::Parser

#print STDERR "\n";

### Verify directory paths:
(-d "testout") or die "missing testout directory\n"
my $output_dir = File::Spec->catdir(".", "testout", "Ref_t")

### Get messages to process:
my @refpaths = @ARGV
if (!nelems(@refpaths))
    opendir my $dh, "testmsgs" or die "opendir: $^OS_ERROR\n"
    @refpaths = map { File::Spec->catfile(".", "testmsgs", $_)
                    }, grep { m/\w.*\.ref$/ }, @: readdir($dh)
    closedir $dh


plan( tests => 2 * scalar(nelems(@refpaths)) )

### For each reference:
foreach my $refpath ( @refpaths)

    ### Get message:
    my $msgpath = $refpath; $msgpath =~ s/\.ref$/.msg/
    # print $^STDERR, "   $msgpath\n";

    ### Get reference, as ref to array:
    my $ref = read_ref($refpath)
    if ($ref->{Parser}->{?Message})
        $msgpath = File::Spec->catfile(".", ( <split m/\//, $ref->{Parser}->{?Message}))
    
    # diag("Trying $refpath [$msgpath]\n");

    ### Create parser which outputs to testout/scratch:
    my $parser = MIME::Parser->new
    $parser->output_dir($output_dir)
    $parser->extract_nested_messages = $ref->{Parser}->{?ExtractNested}
    $parser->extract_uuencode = $ref->{Parser}->{?ExtractUuencode}
    $parser->output_to_core = 0
    $parser->ignore_errors = 0

    ### Set character set:
    my $tgt = $ref->{Parser}->{?Charset} || 'ISO-8859-1'
    my $wd
    if ($tgt =~ m/^ISO-8859-(\d+)/)
        $wd = MIME::WordDecoder::ISO_8859->new( $1)
    else
        $wd = MIME::WordDecoder->new(@: uc($tgt)   => 'KEEP'
                                        'US-ASCII' => 'KEEP'
                                        '*'        => 'WARN')
    
    # diag("Default charset: $tgt");
    MIME::WordDecoder->default($wd)

    ### Pre-clean:
    rmtree($output_dir)
    (-d $output_dir) or mkpath($output_dir) or die "mkpath $output_dir: $^OS_ERROR\n"

    ### Parse:
    my $ent = try { $parser->parse_open($msgpath) }
    my $parse_error = $^EVAL_ERROR

    ### Output parse log:
    #    diag("PARSE LOG FOR $refpath [$msgpath]");
    if ($parser->results) {
    #	diag($parser->results->msgs);
    }else
        diag("Parse failed before results object was created")
    

    ### Interpret results:
    if ($parse_error || !$ent)
        ok($ref->{Msg}->{?Fail}, "$(dump::view($refpath)), problem: $($parse_error && $parse_error->message)" )
    else
        # TODO: check_ref is evil
        my $ok = check_ref($msgpath, $ent, $ref)
        ok($ok, "$(dump::view($refpath)) Message => $msgpath, Parser => " . ($ref->{Parser}->{?Name} || 'default'))
    

    ### Is purge working?
    my @a_files = list_dir($output_dir)
    my @p_files = $parser->filer->purgeable
    $parser->filer->purge
    my @z_files = list_dir($output_dir)
    is(nelems(@z_files), 0, 'Did purge work?')

    ### Cleanup for real:
    rmtree($output_dir)


### Done!
exit(0)
1

#------------------------------

sub list_dir
    my $dir = shift
    opendir my $dh, $dir or die "opendir $dir; $^OS_ERROR\n"
    my @files = grep { !m/^\.+$/ }, @: readdir $dh
    closedir $dh
    return  sort  @files


#------------------------------

sub read_ref
    my $path = shift
    open my $in, "<", "$path" or die "open $path: $^OS_ERROR\n"
    my $expr = join('', @: ~< $in->*)
    close $in
    my $ref = eval $expr; $^EVAL_ERROR and die "syntax error in $path: $($^EVAL_ERROR->message)\n"
    $ref


#------------------------------

sub trim
    local $_ = shift
    s/^\s*//
    s/\s*$//
    $_


#------------------------------
# TODO: replace with cmp_deeply from Test::Deep?
sub check_ref($msgpath, $ent, $ref)

    ### For each Msg in the ref:
    MSG:
        foreach my $partname (sort  keys $ref->%)
        $partname =~ m/^(Msg|Part_)/ or next
        my $msg_ref = $ref->{?$partname}
        my $part    = get_part($ent, $partname) ||
            die "no such part: $partname\n"
        my $head    = $part->head; $head->unfold
        my $body    = $part->bodyhandle

        ### For each attribute in the Msg:
        ATTR:
            foreach (sort  keys $msg_ref->%)

            my $want = $msg_ref->{?$_}
            my $got = undef

            if    (m/^Boundary$/)
                $got = $head->multipart_boundary
            elsif (m/^From$/)
                $got  = trim($head->get("From", 0))
                $want = trim($want)
            elsif (m/^To$/)
                $got  = trim($head->get("To", 0))
                $want = trim($want)
            elsif (m/^Subject$/)
                $got  = trim($head->get("Subject", 0))
                $want = trim($want)
            elsif (m/^Charset$/)
                $got = $head->mime_attr("content-type.charset")
            elsif (m/^Disposition$/)
                $got = $head->mime_attr("content-disposition")
            elsif (m/^Type$/)
                $got = $head->mime_type
            elsif (m/^Encoding$/)
                $got = $head->mime_encoding
            elsif (m/^Filename$/)
                $got = unmime $head->recommended_filename
            elsif (m/^BodyFilename$/)
                $got = (($body and $body->path)
                    ?? basename($body->path)
                    !! undef)
            elsif (m/^Preamble$/)
                $got = join('', $part->preamble->@)
            elsif (m/^Epilogue$/)
                $got = join('', $part->epilogue->@)
            elsif (m/^Size$/)
                if ($head->mime_type =~ m{^(text|message)})
                    #		    diag("Skipping Size evaluation in text message ".
                    #			    "due to variations in local newline ".
                    #			    "conventions\n\n");
                    next ATTR
                
                if ($body and $body->path) { $got = (-s $body->path) }
            else
                die "$partname: unrecognized reference attribute: $_\n"
            

            ### Log this sub-test:
            #	    diag("SUB-TEST: msg=$msgpath; part=$partname; attr=$_:\n");
            #	    diag("  want: ".encode($want)."\n");
            #	    diag("  got:  ".encode($got )."\n");
            #	    diag("\n");

            next ATTR if (!defined($want) and !defined($got))
            next ATTR if ($want eq $got)
            die "$partname: wanted qq\{$want\}, got qq\{$got\}\n"
        
    

    1


# Encode a string
sub encode
    use bytes
    local $_ = shift
    return '<undef>' if !defined($_)

    s{([\n\t\x[00]-\x[1F]\x[7F]-\x[FF]\\\"])}
      {$('\\'.sprintf("\%02X",ord($1)) )}xg
    s{\\0A}{\\n}g
    return qq{"$_"}


#------------------------------

sub get_part($ent, $name)

    if ($name eq 'Msg')
        return $ent
    elsif ($name =~ m/^Part_(.*)$/)
        my @path = split m/_/, $1
        my $part = $ent
        while (nelems(@path))
            my $i = shift @path
            $part = $part->parts($i - 1)
        
        return $part
    
    undef


1

