#!/usr/bin/perl -w


use lib 't/lib'
use MBTest tests => 31
use Module::Build
use Module::Build::ConfigData

my $have_yaml = Module::Build::ConfigData->feature('YAML_support')

#########################

use Cwd ()
my $cwd = Cwd::cwd
my $tmp = File::Spec->catdir( $cwd, 't', '_tmp' )

use DistGen
my $dist = DistGen->new( dir => $tmp )
$dist->remove_file( 't/basic.t' )
$dist->change_file( 'Build.PL', <<'---' )
use Module::Build

my $build = Module::Build->new(
  module_name => 'Simple',
  scripts     => (\@: 'script' ),
  license     => 'perl',
  requires    => (\%: 'File::Spec' => 0 ),
  )
$build->create_build_script
---
$dist->add_file( 'script', <<'---' )
#!perl -w
print $^STDOUT, "Hello, World!\n"
---
$dist->add_file( 'test.pl', <<'---' )
#!/usr/bin/perl

use Test::More
plan tests => 2

ok 1

require Module::Build
like $^INCLUDED{'Module/Build.pm'}, qr/blib/, 'Module::Build should be loaded from blib'

print $^STDOUT, "# Cwd: ", Module::Build->cwd, "\n"
print $^STDOUT, "# \$^INCLUDE_PATH: ($(join ' ', $^INCLUDE_PATH))\n"
print $^STDOUT, "Done.\n"  # t/compat.t looks for this
---
$dist->add_file( 'lib/Simple/Script.PL', <<'---' )
#!perl -w

my $filename = shift(@ARGV)
open my $fh, ">", "$filename" or die "Can't create $filename: $^OS_ERROR"
print $fh, "Contents: $filename\n"
close $fh
---
$dist->regen

chdir( $dist->dirname ) or die "Can't chdir to '$($dist->dirname)': $^OS_ERROR"

#########################

use Module::Build
ok(1)

SKIP: do
    like $^INCLUDED{?'Module/Build.pm'}, qr/\bblib\b/, "Make sure version from blib/ is loaded"


#########################

my $mb = Module::Build->new_from_context
ok $mb
is $mb->license, 'perl'

# Make sure cleanup files added before create_build_script() get respected
$mb->add_to_cleanup('before_script')

try {$mb->create_build_script}
is $^EVAL_ERROR, ''
ok -e $mb->build_script

my $dist_dir = 'Simple-0.01'

# VMS may or may not need to modify the name, vmsify will do this if
# the name looks like a UNIX directory.
if ($^OS_NAME eq 'VMS')
    my @dist_dirs = File::Spec->splitdir( <VMS::Filespec::vmsify($dist_dir.'/'))
    $dist_dir = @dist_dirs[0]


is $mb->dist_dir, $dist_dir

# The 'cleanup' file doesn't exist yet
ok grep {$_ eq 'before_script'}, $mb->cleanup

$mb->add_to_cleanup('save_out')

# The 'cleanup' file now exists
ok grep {$_ eq 'before_script'}, $mb->cleanup
ok grep {$_ eq 'save_out'     }, $mb->cleanup

do
    # Make sure verbose=>1 works
    my $all_ok = 1
    my $output = stdout_of( sub (@< @_) { $mb->dispatch('test', verbose => 1) } )
    $all_ok &&= like($output, qr/all tests successful/i)

    # This is the output of lib/Simple/Script.PL
    $all_ok &&= ok(-e $mb->localize_file_path('lib/Simple/Script'))

    unless ($all_ok)
        # We use diag() so Test::Harness doesn't get confused.
        diag("vvvvvvvvvvvvvvvvvvvvv Simple/test.pl output vvvvvvvvvvvvvvvvvvvvv")
        diag($output)
        diag("^^^^^^^^^^^^^^^^^^^^^ Simple/test.pl output ^^^^^^^^^^^^^^^^^^^^^")
    


SKIP: do
    skip( 'YAML_support feature is not enabled', 7 ) unless $have_yaml

    my $output = try {
        stdout_of( sub (@< @_) { $mb->dispatch('disttest') } )
    }
    is $^EVAL_ERROR, ''

    # After a test, the distdir should contain a blib/ directory
    ok -e File::Spec->catdir('Simple-0.01', 'blib')

    try {$mb->dispatch('distdir')}
    is $^EVAL_ERROR && $^EVAL_ERROR->message, ''

    # The 'distdir' should contain a lib/ directory
    ok -e File::Spec->catdir('Simple-0.01', 'lib')

    # The freshly run 'distdir' should never contain a blib/ directory, or
    # else it could get into the tarball
    ok ! -e File::Spec->catdir('Simple-0.01', 'blib')

    # Make sure all of the above was done by the new version of Module::Build
    my $fh = IO::File->new( <File::Spec->catfile( <$dist->dirname, 'META.yml'))
    my $contents = do {local $^INPUT_RECORD_SEPARATOR = undef; ~< $fh}
    $contents =~ m/Module::Build version ([0-9_.]+)/m
    cmp_ok $1, '==', $mb->VERSION, "Check version used to create META.yml: $1 == " . $mb->VERSION

    SKIP: do
        skip( "not sure if we can create a tarball on this platform", 1 )
            unless $mb->check_installed_status('Archive::Tar', 0) ||
            $mb->isa('Module::Build::Platform::Unix')

        $mb->add_to_cleanup($mb->dist_dir . ".tar.gz")
        try {$mb->dispatch('dist')}
        is $^EVAL_ERROR, ''
    



do
    # Make sure the 'script' file was recognized as a script.
    my $scripts = $mb->script_files
    ok $scripts->{?script}

    # Check that a shebang line is rewritten
    my $blib_script = File::Spec->catfile( < qw( blib script script ) )
    ok -e $blib_script

    SKIP: do
        skip("We do not rewrite shebang on VMS", 1) if $^OS_NAME eq 'VMS'
        my $fh = IO::File->new($blib_script)
        my $first_line = ~< $fh
        isnt $first_line, "#!perl -w\n", "should rewrite the shebang line"
    


do
    # Check PPD
    $mb->dispatch('ppd', args => \(%: codebase => '/path/to/codebase'))

    my $ppd = slurp('Simple.ppd')

    # This test is quite a hack since with XML you don't really want to
    # do a strict string comparison, but absent an XML parser it's the
    # best we can do.
    is $ppd, <<'EOF'
<SOFTPKG NAME="Simple" VERSION="0,01,0,0">
    <TITLE>Simple</TITLE>
    <ABSTRACT>Perl extension for blah blah blah</ABSTRACT>
    <AUTHOR>A. U. Thor, a.u.thor@a.galaxy.far.far.away</AUTHOR>
    <IMPLEMENTATION>
        <DEPENDENCY NAME="File-Spec" VERSION="0,0,0,0" />
        <CODEBASE HREF="/path/to/codebase" />
    </IMPLEMENTATION>
</SOFTPKG>
EOF



try {$mb->dispatch('realclean')}
is $^EVAL_ERROR, ''

ok ! -e $mb->build_script
ok ! -e $mb->config_dir
ok ! -e $mb->dist_dir

chdir( $cwd ) or die "Can''t chdir to '$cwd': $^OS_ERROR"
$dist->remove

SKIP:
    do
    skip( 'Windows-only test', 4 ) unless $^OS_NAME =~ m/^MSWin/

    my $script_data = <<'---'
@echo off
echo Hello, World!
---

    $dist = DistGen->new( dir => $tmp )
    $dist->change_file( 'Build.PL', <<'---' )
use Module::Build;
my $build = new Module::Build(
  module_name => 'Simple',
  scripts     => (\@: 'bin/script.bat' ),
  license     => 'perl',
);
$build->create_build_script;
---
    $dist->add_file( 'bin/script.bat', $script_data )

    $dist->regen
    chdir( $dist->dirname )
        or die "Can't chdir to '$($dist->dirname)': $^OS_ERROR"

    $mb = Module::Build->new_from_context
    ok $mb

    try{ $mb->dispatch('build') }
    is $^EVAL_ERROR, ''

    my $script_file = File::Spec->catfile( < qw(blib script), 'script.bat' )
    ok -f $script_file, "Native batch file copied to 'scripts'"

    my $out = slurp( $script_file )
    is $out, $script_data, '  unmodified by pl2bat'

    chdir( $cwd ) or die "Can''t chdir to '$cwd': $^OS_ERROR"
    $dist->remove


# cleanup
chdir( $cwd ) or die "Can''t chdir to '$cwd': $^OS_ERROR"
$dist->remove

use File::Path
rmtree( $tmp )
