package PPI::Token

=pod

=head1 NAME

PPI::Token - A single token of Perl source code

=head1 INHERITANCE

  PPI::Token
  isa PPI::Element

=head1 DESCRIPTION

C<PPI::Token> is the abstract base class for all Tokens. In PPI terms, a "Token" is
a L<PPI::Element> that directly represents bytes of source code.

=head1 METHODS

=cut

use base 'PPI::Element'
use Params::Util '_INSTANCE'
use PPI::Exception ()

our ($VERSION)
BEGIN 
    $VERSION = '1.204_01'


# We don't load the abstracts, they are loaded
# as part of the 'use base' statements.

# Load the token classes
require PPI::Token::Whitespace
require PPI::Token::Comment
require PPI::Token::Pod
require PPI::Token::Number
require PPI::Token::Number::Binary
require PPI::Token::Number::Octal
require PPI::Token::Number::Hex
require PPI::Token::Number::Float
require PPI::Token::Number::Exp
require PPI::Token::Number::Version
require PPI::Token::Word
require PPI::Token::DashedWord
require PPI::Token::Symbol
require PPI::Token::ArrayIndex
require PPI::Token::Magic
require PPI::Token::Quote::Single
require PPI::Token::Quote::Double
require PPI::Token::Quote::Literal
require PPI::Token::Quote::Interpolate
require PPI::Token::QuoteLike::Backtick
require PPI::Token::QuoteLike::Command
require PPI::Token::QuoteLike::Regexp
require PPI::Token::QuoteLike::Words
require PPI::Token::QuoteLike::Readline
require PPI::Token::Regexp::Match
require PPI::Token::Regexp::Substitute
require PPI::Token::Regexp::Transliterate
require PPI::Token::Operator
require PPI::Token::Cast
require PPI::Token::Structure
require PPI::Token::Label
require PPI::Token::HereDoc
require PPI::Token::Separator
require PPI::Token::Data
require PPI::Token::End
require PPI::Token::Prototype
require PPI::Token::Attribute
require PPI::Token::Unknown





#####################################################################
# Constructor and Related

sub new
    bless \(%:  content => (defined @_[1] ?? "@_[1]" !! '') ), @_[0]


sub set_class
    my $self  = shift
    # @_ or throw Exception("No arguments to set_class");
    my $class = substr( @_[0], 0, 12 ) eq 'PPI::Token::' ?? shift !! 'PPI::Token::' . shift

    # Find out if the current and new classes are complex
    my $old_quote = (ref($self) =~ m/\b(?:Quote|Regex)\b/o) ?? 1 !! 0
    my $new_quote = ($class =~ m/\b(?:Quote|Regex)\b/o)     ?? 1 !! 0

    # No matter what happens, we will have to rebless
    bless $self, $class

    # If we are changing to or from a Quote style token, we
    # can't just rebless and need to do some extra thing
    # Otherwise, we have done enough
    return $class if ($old_quote - $new_quote) == 0

    # Make a new token from the old content, and overwrite the current
    # token's attributes with the new token's attributes.
    my $token = $class->new( $self->{content} )
    $self->% = (%:  < $token->% )

    # Return the class as a convenience
    return $class






#####################################################################
# PPI::Token Methods

=pod

=head2 set_content $string

The C<set_content> method allows to set/change the string that the
C<PPI::Token> object represents.

Returns the string you set the Token to

=cut

sub set_content
    @_[0]->{+content} = @_[1]


=pod

=head2 add_content $string

The C<add_content> method allows you to add additional bytes of code
to the end of the Token.

Returns the new full string after the bytes have been added.

=cut

sub add_content { @_[0]->{+content} .= @_[1] }

=pod

=head2 length

The C<length> method returns the length of the string in a Token.

=cut

sub length { CORE::length(@_[0]->{?content}) }





#####################################################################
# Overloaded PPI::Element methods

sub content
    @_[0]->{?content}


# You can insert either a statement, or a non-significant token.
sub insert_before
    my $self    = shift
    my $Element = _INSTANCE(shift, 'PPI::Element')  or return undef
    if ( $Element->isa('PPI::Structure') )
        return $self->__insert_before($Element)
    elsif ( $Element->isa('PPI::Token') )
        return $self->__insert_before($Element)
    
    ''


# As above, you can insert a statement, or a non-significant token
sub insert_after
    my $self    = shift
    my $Element = _INSTANCE(shift, 'PPI::Element') or return undef
    if ( $Element->isa('PPI::Structure') )
        return $self->__insert_after($Element)
    elsif ( $Element->isa('PPI::Token') )
        return $self->__insert_after($Element)
    
    ''






#####################################################################
# Tokenizer Methods

sub __TOKENIZER__on_line_start { 1 }
sub __TOKENIZER__on_line_end   { 1 }
sub __TOKENIZER__on_char       { 'Unknown' }





#####################################################################
# Lexer Methods

sub __LEXER__opens
    ref(@_[0]) eq 'PPI::Token::Structure'
        and
        @_[0]->{?content} =~ m/(?:\(|\[|\{)/


sub __LEXER__closes
    ref(@_[0]) eq 'PPI::Token::Structure'
        and
        @_[0]->{?content} =~ m/(?:\)|\]|\})/


1
