package t::lib::PPI

use File::Spec::Functions ':ALL'
use Test::More
use Test::Object
use Params::Util    '_INSTANCE'
use List::MoreUtils 'any'
use PPI::Dumper

our ($VERSION)
BEGIN 
    $VERSION = '1.204'






#####################################################################
# PPI::Document Testing

Test::Object->register(
    class => 'PPI::Document',
    tests => 1,
    code  => &document_ok,
    )

sub document_ok
    my $doc = shift

    # A document should have zero or more children that are either
    # a statement or a non-significant child.
    my @children = $doc->children
    my $good = grep {
            _INSTANCE($_, 'PPI::Statement')
                or (
                    _INSTANCE($_, 'PPI::Token') and ! $_->significant
                    )
        }, @children

    is( nelems($good), nelems(@children),
     'Document contains only statements and non-significant tokens' )

    1






#####################################################################
# Are there an unknowns

Test::Object->register(
    class => 'PPI::Document',
    tests => 3,
    code  => &unknown_objects,
    )

sub unknown_objects
    my $doc = shift

    is(
       $doc->find_any('Token::Unknown'),
       '',
       "Contains no PPI::Token::Unknown elements",
       )
    is(
       $doc->find_any('Structure::Unknown'),
       '',
       "Contains no PPI::Structure::Unknown elements",
       )
    is(
       $doc->find_any('Statement::Unknown'),
       '',
       "Contains no PPI::Statement::Unknown elements",
       )

    1





#####################################################################
# Are there any invalid nestings?

Test::Object->register(
    class => 'PPI::Document',
    tests => 1,
    code  => &nested_statements,
    )

sub nested_statements
    my $doc = shift

    ok(
       ! $doc->find_any( sub (@< @_)
                             _INSTANCE(@_[1], 'PPI::Statement')
                                 and any { _INSTANCE($_, 'PPI::Statement') }, < @_[1]->children
                        ),
       'Document contains no nested statements',
       )


Test::Object->register(
    class => 'PPI::Document',
    tests => 1,
    code  => &nested_structures,
    )

sub nested_structures
    my $doc = shift

    ok(
       ! $doc->find_any( sub (@< @_)
                             _INSTANCE(@_[1], 'PPI::Structure')
                                 and any { _INSTANCE($_, 'PPI::Structure') }, < @_[1]->children
                       ),
       'Document contains no nested structures',
       )


Test::Object->register(
    class => 'PPI::Document',
    tests => 1,
    code  => &no_attribute_in_attribute,
    )

sub no_attribute_in_attribute
    my $doc = shift

    ok(
       ! $doc->find_any( sub (@< @_)
                             _INSTANCE(@_[1], 'PPI::Token::Attribute')
                                 and ! exists @_[1]->{_attribute}
                       ),
       'No ->{_attribute} in PPI::Token::Attributes',
       )






#####################################################################
# Process a .code/.dump file pair
# plan: 2 + 14 * npairs

sub run_testdir
    my $pkg     = shift
    my $testdir = catdir(< @_)

    # Does the test directory exist?
    ok( (-e $testdir and -d $testdir and -r $testdir), "Test directory $testdir found" )

    # Find the .code test files
    opendir( my $testdh, $testdir ) or die "opendir: $^OS_ERROR"
    my @code = map { catfile( $testdir, $_ ) }, sort grep { m/\.code$/ }, @:  readdir($testdh) 
    closedir( $testdh ) or die "closedir: $^OS_ERROR"
    ok( scalar nelems(@code), 'Found at least one code file' )

    foreach my $codefile ( @code )
        # Does the .code file have a matching .dump file
        my $dumpfile = $codefile
        $dumpfile =~ s/\.code$/\.dump/
        my $codename = $codefile
        $codename =~ s/\.code$//
        ok( (-f $dumpfile and -r $dumpfile), "$codename: Found matching .dump file" )

        # Create the lexer and get the Document object
        my $Document = PPI::Document->new( $codefile )
        ok( $Document, "$codename: Lexer->Document returns true" )
        ok( _INSTANCE($Document, 'PPI::Document'), "$codename: Object isa PPI::Document" )

        my $rv
        my $code_fh
        SKIP: do
            skip "No Document to test", 11 unless $Document

            # Check standard things
            object_ok( $Document ) # 7 tests contained within

            # Get the dump array ref for the Document object
            my $Dumper = PPI::Dumper->new( $Document )
            ok( _INSTANCE($Dumper, 'PPI::Dumper'), "$codename: Object isa PPI::Dumper" )
            my @dump_list = $Dumper->list
            ok( nelems(@dump_list), "$codename: Got dump content from dumper" )

            # Try to get the .dump file array
            open( my $dumpfh, '<', $dumpfile ) or die "open: $^OS_ERROR"
            my @content = @:  ~< $dumpfh->* 
            close( $dumpfh ) or die "close: $^OS_ERROR"
            chomp @content

            # Compare the two
            is_deeply( \@dump_list, \@content, "$codename: Generated dump matches stored dump" )

            # Also, do a round-trip check
            $rv = open( $code_fh, '<', $codefile )
            ok( $rv, "$codename: Opened file" )
        
        SKIP: do
            unless ( $Document and $rv )
                skip "Missing file", 1
            
            my $source = do { local $^INPUT_RECORD_SEPARATOR = undef; ~< $code_fh }
            close $code_fh
            $source =~ s/(?:\015{1,2}\012|\015|\012)/\n/g

            is( $Document->serialize, $source, "$codename: Round-trip back to source was ok" )
        
    







#####################################################################
# Process a .code/.dump file pair
# plan: 2 + 14 * npairs

sub increment_testdir
    my $pkg     = shift
    my $testdir = catdir(< @_)

    # Does the test directory exist?
    ok( (-e $testdir and -d $testdir and -r $testdir), "Test directory $testdir found" )

    # Find the .code test files
    opendir( my $testdh, $testdir ) or die "opendir: $^OS_ERROR"
    my @code = map { catfile( $testdir, $_ ) }, sort grep { m/\.code$/ }, @:  readdir($testdh) 
    closedir( $testdh ) or die "closedir: $^OS_ERROR"
    ok( scalar nelems(@code), 'Found at least one code file' )

    foreach my $codefile ( @code )
        # Does the .code file have a matching .dump file
        my $codename = $codefile
        $codename =~ s/\.code$//

        # Load the file
        local $^INPUT_RECORD_SEPARATOR = undef
        open( my $code_fh, "<", $codefile ) or die "open: $^OS_ERROR"
        my $buffer = ~< $code_fh->*
        close( $code_fh ) or die "close: $^OS_ERROR"

        # Cover every possible transitional state in
        # the regression test code fragments.
        foreach my $chars ( 1 .. length($buffer) )
            my $string   = substr( $buffer, 0, $chars )
            my $Document = PPI::Document->new( \$string )
            is(
               ref($Document) => 'PPI::Document',
               "$codename: $chars chars document",
               )
            is(
               $Document->serialize => $string,
               "$codename: $chars char roundtrip",
               )


1
