package TAP::Parser::Iterator::Process


require TAP::Parser::Iterator

our ($VERSION, @ISA)

@ISA = @:  'TAP::Parser::Iterator' 

use Config
use IO::Handle

my $IS_WIN32 = ( $^OS_NAME =~ m/^(MS)?Win32$/ )

=head1 NAME

TAP::Parser::Iterator::Process - Internal TAP::Parser Iterator

=head1 VERSION

Version 3.10

=cut

$VERSION = '3.10'

=head1 SYNOPSIS

  use TAP::Parser::Iterator;
  my $it = TAP::Parser::Iterator::Process->new(@args);

  my $line = $it->next;

Originally ripped off from L<Test::Harness>.

=head1 DESCRIPTION

B<FOR INTERNAL USE ONLY!>

This is a simple iterator wrapper for processes.

=head2 Class Methods

=head3 C<new>

Create an iterator.

=head2 Instance Methods

=head3 C<next>

Iterate through it, of course.

=head3 C<next_raw>

Iterate raw input without applying any fixes for quirky input syntax.

=head3 C<wait>

Get the wait status for this iterator's process.

=head3 C<exit>

Get the exit status for this iterator's process.

=cut

try { require POSIX; POSIX::WEXITSTATUS(0) }
if ($^EVAL_ERROR)
    *_wait2exit = sub (@< @_) { @_[1] >> 8 }
else
    *_wait2exit = sub (@< @_) { POSIX::WEXITSTATUS( @_[1] ) }


sub _use_open3
    my $self = shift
    return unless config_value("d_fork") || $IS_WIN32
    for my $module (qw( IPC::Open3 IO::Select ))
        eval "use $module"
        return if $^EVAL_ERROR
    
    return 1


do
    my $got_unicode

    sub _get_unicode
        return $got_unicode if defined $got_unicode
        eval 'use Encode qw(decode_utf8);'
        $got_unicode = $^EVAL_ERROR ?? 0 !! 1

    


sub new
    my $class = shift
    my $args  = shift

    my @command = ( delete $args->{command} || \$@ )->@
        or die "Must supply a command to execute"

    # Private. Used to frig with chunk size during testing.
    my $chunk_size = delete $args->{_chunk_size} || 65536

    my $merge = delete $args->{merge}
    my ( $pid, $err, $sel )

    if ( my $setup = delete $args->{setup} )
        $setup->(< @command)
    

    my $out = IO::Handle->new

    if ( $class->_use_open3 )

        # HOTPATCH {{{
        my $xclose = &IPC::Open3::xclose
        local $^WARNING = 0    # no warnings
        local *IPC::Open3::xclose = sub (@< @_)
            my $fh = shift

            return if ( fileno($fh) == fileno($^STDIN) )
            $xclose->($fh)
        

        # }}}

        if ($IS_WIN32)
            $err = $merge ?? '' !! (@: '>&', $^STDERR)
            try {
                $pid = open3(
                    (@: '<&', $^STDIN), $out, $merge ?? '' !! $err,
                    < @command
                    );
            }
            die "Could not execute ($(join ' ',@command)): $^EVAL_ERROR" if $^EVAL_ERROR
            binmode($out, ":crlf")
        else
            $err = $merge ?? '' !! IO::Handle->new
            try { $pid = open3( $^STDIN, $out, $err, < @command ); }
            die "Could not execute ($(join ' ',@command)): $($^EVAL_ERROR->message)" if $^EVAL_ERROR
            $sel = $merge ?? undef !! IO::Select->new( $out, $err )
        
    else
        $err = ''
        my $command
            = join( ' ', map { $_ =~ m/\s/ ?? qq{"$_"} !! $_ }, @command )
        open( $out, "-|", $command )
            or die "Could not execute ($command): $^OS_ERROR"
    

    my $self = bless \(%: 
        out        => $out
        err        => $err
        sel        => $sel
        pid        => $pid
        exit       => undef
        chunk_size => $chunk_size
        ), $class

    if ( my $teardown = delete $args->{teardown} )
        $self->{+teardown} = sub (@< @_)
            $teardown->(< @command)
        
    

    return $self


=head3 C<handle_unicode>

Upgrade the input stream to handle UTF8.

=cut

sub handle_unicode
    my $self = shift

    if ( $self->{?sel} )
        if ( _get_unicode() )

            # Make sure our iterator has been constructed and...
            my $next = $: $self->{+_next} ||= $self->_next

            # ...wrap it to do UTF8 casting
            $self->{+_next} = sub (@< @_)
                my $line = $next->()
                return decode_utf8($line) if defined $line
                return
            
        
    else
        binmode($self->{?out}, ":utf8")
    



##############################################################################

sub wait { shift->{?wait} }
sub exit { shift->{?exit} }

sub _next
    my $self = shift

    if ( my $out = $self->{?out} )
        if ( my $sel = $self->{?sel} )
            my $err        = $self->{?err}
            my @buf        = $@
            my $partial    = ''                    # Partial line
            my $chunk_size = $self->{?chunk_size}
            return sub (@< @_)
                return shift @buf if (nelems @buf)

                READ:
                    while ( my @ready = $sel->can_read )
                    for my $fh ( @ready)
                        my $got = sysread $fh, my ($chunk), $chunk_size

                        if ( $got == 0 )
                            $sel->remove($fh)
                        elsif ( $fh \== $err )
                            print $^STDERR, $chunk    # echo STDERR
                        else
                            $chunk   = $partial . $chunk
                            $partial = ''

                            # Make sure we have a complete line
                            unless ( substr( $chunk, -1, 1 ) eq "\n" )
                                my $nl = rindex $chunk, "\n"
                                if ( $nl == -1 )
                                    $partial = $chunk
                                    redo READ
                                else
                                    $partial = substr( $chunk, $nl + 1 )
                                    $chunk = substr( $chunk, 0, $nl )
                                
                            

                            push @buf, < split m/\n/, $chunk
                            return shift @buf if (nelems @buf)
                        
                    
                

                # Return partial last line
                if ( length $partial )
                    my $last = $partial
                    $partial = ''
                    return $last
                

                $self->_finish
                return
            
        else
            return sub (@< @_)
                if ( defined( my $line = ~< $out ) )
                    chomp $line
                    return $line
                
                $self->_finish
                return
            
        
    else
        return sub (@< @_)
            $self->_finish
            return
        
    


sub next_raw
    my $self = shift
    return ( $self->{+_next} ||= $self->_next )->()


sub _finish
    my $self = shift

    my $status = $^CHILD_ERROR

    # If we have a subprocess we need to wait for it to terminate
    if ( defined $self->{?pid} )
        if ( $self->{?pid} == waitpid( $self->{?pid}, 0 ) )
            $status = $^CHILD_ERROR
        
    

    ( delete $self->{out} )->close if $self->{?out}

    # If we have an IO::Select we also have an error handle to close.
    if ( $self->{?sel} )
        ( delete $self->{err} )->close
        delete $self->{sel}
    else
        $status = $^CHILD_ERROR
    

    # Sometimes we get -1 on Windows. Presumably that means status not
    # available.
    $status = 0 if $IS_WIN32 && $status == -1

    $self->{+wait} = $status
    $self->{+exit} = $self->_wait2exit($status)

    if ( my $teardown = $self->{?teardown} )
        $teardown->()
    

    return $self


=head3 C<get_select_handles>

Return a list of filehandles that may be used upstream in a select()
call to signal that this Iterator is ready. Iterators that are not
handle based should return an empty list.

=cut

sub get_select_handles
    my $self = shift
    return grep { $_ }, @:  ( $self->{?out}, $self->{?err} )


1
