package TAP::Parser::Source::Perl

use Config
our ($VERSION, @ISA)

use constant IS_WIN32 => ( $^OS_NAME =~ m/^(MS)?Win32$/ )
use constant IS_VMS => ( $^OS_NAME eq 'VMS' )

use TAP::Parser::Source
@ISA = @: 'TAP::Parser::Source'

=head1 NAME

TAP::Parser::Source::Perl - Stream Perl output

=head1 VERSION

Version 3.10

=cut

$VERSION = '3.10'

=head1 DESCRIPTION

Takes a filename and hopefully returns a stream from it.  The filename should
be the name of a Perl program.

Note that this is a subclass of L<TAP::Parser::Source>.  See that module for
more methods.

=head1 SYNOPSIS

    use TAP::Parser::Source::Perl;
    my $perl = TAP::Parser::Source::Perl->new;
    my $stream = $perl->source( [ $filename, @args ] )->get_stream;

=head1 METHODS

=head2 Class Methods

=head3 C<new>

 my $perl = TAP::Parser::Source::Perl->new;

Returns a new C<TAP::Parser::Source::Perl> object.

=head2 Instance Methods

=head3 C<source>

Getter/setter the name of the test program and any arguments it requires.

  my ($filename, @args) = @{ $perl->source };
  $perl->source( [ $filename, @args ] );

=cut

sub source
    my $self = shift
    $self->_croak("Cannot find (@_[0]->[0])")
        if (nelems @_) && !-f @_[0]->[0]
    return $self->SUPER::source(< @_)


=head3 C<switches>

  my $switches = $perl->switches;
  my @switches = $perl->switches;
  $perl->switches = @switches;

Getter/setter for the additional switches to pass to the perl executable.  One
common switch would be to set an include directory:

  $perl->switches = @: '-Ilib';

=cut

sub switches($self ?= $value)
    if ($^is_assignment)
        $self->{+switches} = $value
    
    return  $self->{?switches}


##############################################################################

=head3 C<get_stream>

 my $stream = $source->get_stream;

Returns a stream of the output generated by executing C<source>.

=cut

sub get_stream
    my $self = shift

    my @extra_libs

    my @switches = $self->_switches
    my $path_sep = config_value("path_sep")
    my $path_pat = qr{$path_sep}

    # Nasty kludge. It might be nicer if we got the libs separately
    # although at least this way we find any -I switches that were
    # supplied other then as explicit libs.
    # We filter out any names containing colons because they will break
    # PERL5LIB
    my @libs
    for ( grep { $_ !~ $path_pat }, @switches )
        push @libs, $1 if m/ ^ ['"]? -I (.*?) ['"]? $ /x
    

    my $previous = env::var('PERL5LIB')
    if ($previous)
        push @libs, < split( $path_pat, $previous )
    

    my $setup = sub (@< @_)
        if ((nelems @libs))
            env::var('PERL5LIB' ) = join( $path_sep, @libs )
        
    

    # Cargo culted from comments seen elsewhere about VMS / environment
    # variables. I don't know if this is actually necessary.
    my $teardown = sub (@< @_)
        env::var('PERL5LIB' ) = $previous
    

    # Taint mode ignores environment variables so we must retranslate
    # PERL5LIB as -I switches and place PERL5OPT on the command line
    # in order that it be seen.
    if ( grep { $_ eq "-T" }, @switches )
        push @switches, <
            $self->_libs2switches( <
            split $path_pat,
            env::var('PERL5LIB') || env::var('PERLLIB') || ''
            )

        push @switches, env::var('PERL5OPT') || ()
    

    my @command = $self->_get_command_for_switches(< @switches)
        or $self->_croak("No command found!")

    return TAP::Parser::Iterator->new(
        \%:  command  => \@command
             merge    => $self->merge
             setup    => $setup
             teardown => $teardown
        
        )


sub _get_command_for_switches($self, @< @switches)
    my @: $file, @< @args = $self->source->@
    my $command = $self->_get_perl

    $file = qq["$file"] if ( $file =~ m/\s/ ) && ( $file !~ m/^".*"$/ )
    my @command = @: $command, < @switches, $file, < @args
    return @command


sub _get_command($self)
    return $self->_get_command_for_switches( < $self->_switches )


sub _libs2switches
    my $self = shift
    return map {"-I$_"}, grep {$_}, @_


=head3 C<shebang>

Get the shebang line for a script file.

    my $shebang = TAP::Parser::Source::Perl->shebang( $some_script );

May be called as a class method

=cut

do

    # Global shebang cache.
    my %shebang_for

    sub _read_shebang
        my $file = shift
        my $test
        my $shebang
        if ( open( $test, "<", $file ) )
            $shebang = ~< $test
            close($test) or print $^STDOUT, "Can't close $file. $^OS_ERROR\n"
        else
            print $^STDOUT, "Can't open $file. $^OS_ERROR\n"
        
        return $shebang
    

    sub shebang( $class, $file)
        unless ( exists %shebang_for{$file} )
            %shebang_for{+$file} = _read_shebang($file)
        
        return %shebang_for{?$file}
    


sub _switches($self)
    my @: $file, @< @args =   $self->source->@
    my @switches = $self->switches

    my $shebang = $self->shebang($file)
    return unless defined $shebang

    # Quote the argument if there's any whitespace in it, or if
    # we're VMS, since VMS requires all parms quoted.  Also, don't quote
    # it if it's already quoted.
    for (@switches)
        $_ = qq["$_"] if ( ( m/\s/ || IS_VMS ) && ! m/^".*"$/ )

    return @switches


sub _get_perl
    my $proto = shift
    return env::var('HARNESS_PERL') if defined env::var('HARNESS_PERL')
    return Win32::GetShortPathName($^EXECUTABLE_NAME) if IS_WIN32
    return $^EXECUTABLE_NAME


1
