#!/usr/bin/perl -w


use lib 't/lib'

use Test::More tests => 37

use TAP::Base

do

    # No callbacks allowed
    can_ok 'TAP::Base', 'new'
    my $base = TAP::Base->new()
    isa_ok $base, 'TAP::Base', 'object of correct type'
    foreach my $method (qw(callback _callback_for _initialize))
        can_ok $base, $method
    

    try {
        $base->callback(
            some_event => sub {

            # do nothing
            }
            );
    }
    like( $^EVAL_ERROR->message, qr/No callbacks/, 'no callbacks allowed croaks OK' )
    my $cb = $base->_callback_for('some_event')
    ok( !$cb, 'no callback installed' )


do

    # No callbacks allowed, constructor should croak
    try {
        my $base = TAP::Base->new(
            %:    callbacks => \%: 
                      some_event => sub {

                      # do nothing
                      }
            
            
            );
    }
    like(
        $^EVAL_ERROR->message, qr/No callbacks/,
        'no callbacks in constructor croaks OK'
        )


package CallbackOK

use TAP::Base
our (@ISA)
@ISA = @:  'TAP::Base' 

sub _initialize
    my $self = shift
    my $args = shift
    $self->SUPER::_initialize( $args, \qw( nice_event other_event ) )
    return $self


package main
do
    my $base = CallbackOK->new()
    isa_ok $base, 'TAP::Base'

    try {
        $base->callback(
            some_event => sub {

            # do nothing
            }
            );
    }
    like( $^EVAL_ERROR->message, qr/Callback some_event/, 'illegal callback croaks OK' )

    my (@:  $nice, $other ) = @:  0, 0 

    $base->callback( other_event => sub (@< @_) { $other-- } )
    $base->callback( nice_event => sub (@< @_) { $nice++; return shift() . 'OK' } )

    ok( 1, 'callbacks installed OK' )

    my $nice_cbs = $base->_callback_for('nice_event')
    is( ref::svtype($nice_cbs), 'ARRAY', 'callbacks type ok' )
    is( nelems $nice_cbs, 1, 'right number of callbacks' )
    my $nice_cb = $nice_cbs[0]
    ok( type::is_code($nice_cb), 'callback for nice_event returned' )
    my $got = $nice_cb->('Is ')
    is( $got, 'Is OK', 'args passed to callback' )
    cmp_ok( $nice, '==', 1, 'callback calls the right sub' )

    my $other_cbs = $base->_callback_for('other_event')
    is( ref::svtype($other_cbs), 'ARRAY', 'callbacks type ok' )
    is( nelems $other_cbs, 1, 'right number of callbacks' )
    my $other_cb = $other_cbs[0]
    ok( type::is_code($other_cb), 'callback for other_event returned' )
    $other_cb->()
    cmp_ok( $other, '==', -1, 'callback calls the right sub' )

    my @got = $base->_make_callback( 'nice_event', 'I am ' )
    is( scalar nelems @got, 1, 'right number of results' )
    is( @got[0], 'I am OK', 'callback via _make_callback works' )


do
    my (@:  $nice, $other ) = @:  0, 0 

    my $base = CallbackOK->new(
        %:    callbacks => \%: 
                  nice_event => sub (@< @_) { $nice++ }
        
        
        )

    isa_ok $base, 'TAP::Base', 'object creation with callback succeeds'

    try {
        $base->callback(
            some_event => sub {

            # do nothing
            }
            );
    }
    like( $^EVAL_ERROR->message, qr/Callback some_event/, 'illegal callback croaks OK' )

    try {
        $base->callback( other_event => sub (@< @_) { $other-- } );
    }

    ok( !$^EVAL_ERROR, 'callback installed OK' )

    my $nice_cbs = $base->_callback_for('nice_event')
    is( ref::svtype($nice_cbs), 'ARRAY', 'callbacks type ok' )
    is( nelems $nice_cbs, 1, 'right number of callbacks' )
    my $nice_cb = $nice_cbs[0]
    ok( type::is_code($nice_cb), 'callback for nice_event returned' )
    $nice_cb->()
    cmp_ok( $nice, '==', 1, 'callback calls the right sub' )

    my $other_cbs = $base->_callback_for('other_event')
    is( ref::svtype($other_cbs), 'ARRAY', 'callbacks type ok' )
    is( nelems $other_cbs, 1, 'right number of callbacks' )
    my $other_cb = $other_cbs[0]
    ok( type::is_code($other_cb), 'callback for other_event returned' )
    $other_cb->()
    cmp_ok( $other, '==', -1, 'callback calls the right sub' )

    # my @got = $base->_make_callback( 'nice_event', 'I am ' );
    # is ( scalar @got, 1, 'right number of results' );
    # is( $got[0], 'I am OK', 'callback via _make_callback works' );

    my $status = undef

    # Stack another callback
    $base->callback( other_event => sub (@< @_) { $status = 'OK'; return 'Aye' } )

    my $new_cbs = $base->_callback_for('other_event')
    is( ref::svtype($new_cbs), 'ARRAY', 'callbacks type ok' )
    is( nelems $new_cbs, 2, 'right number of callbacks' )
    my $new_cb = $new_cbs[1]
    ok( type::is_code($new_cb), 'callback for new_event returned' )
    my @got = @:  $new_cb->() 
    is( $status, 'OK', 'new callback called OK' )

