#!/usr/local/bin/perl -w


use Test::More 
use Text::Diff 
use Algorithm::Diff < qw( traverse_sequences ) 

## Each test specifies options to pass to "diff" when the --update option
## is present in @ARGV and options to pass to Text::Diff::diff when the
## tests are run.
my @tests= @: 
    \(@: "-u"
         ''
      )
    \(@: "-c"
         'STYLE => "Context"'
      )
    \(@: "-C0"
         'STYLE => "Context", CONTEXT => 0'
      )
    \(@: "-U0"
         'STYLE => "Unified", CONTEXT => 0'
      )
    \(@: ""
         'STYLE => "OldStyle"'
      )
     

my @A = map { "$_\n" }, qw( 1 2 3 4 5d 6 7 8 9    10 11 11d 12 13 ) 
my @B = map { "$_\n" }, qw( 1 2 3 4 5a 6 7 8 9 9a 10 11     12 13 ) 

my $sep = ( "----8<----" x 7 ) . "\n" 

if ( grep { "--update" }, @ARGV )
    my $version = `diff -v` 

    die "Could not determine your diff's version"
        unless defined $version && length $version 
    chomp $version 
    die "Requires GNU's diff, not '$version'"
        unless $version =~ m/GNU/ 

    ## Here are the two files to feed to diff
    open my $a_fh, ">", "A" or die $^OS_ERROR ; print $a_fh, < @A ; close $a_fh 
    open my $b_fh, ">", "B" or die $^OS_ERROR ; print $b_fh, < @B ; close $b_fh 

    my $mtime_A = time 
    my $mtime_B = $mtime_A + 1 

    utime $mtime_A, $mtime_A, "A" or die $^OS_ERROR 
    utime $mtime_B, $mtime_B, "B" or die $^OS_ERROR 

    my $file_options = <<END_OPTIONS 
FILENAME_A => "A",
MTIME_A => $mtime_A,
FILENAME_B => "B",
MTIME_B => $mtime_B,
END_OPTIONS

    open my $me_fh, "<", "$^PROGRAM_NAME" or die $^OS_ERROR 
    my $me = join( "", (@:  ~< $me_fh->*) ) 
    close $me_fh or die $^OS_ERROR 

    open my $bak_fh, ">", "$^PROGRAM_NAME.bak" or die $^OS_ERROR 
    print $bak_fh, $me or die $^OS_ERROR 
    close $bak_fh or die $^OS_ERROR 

    my @diffs = map { scalar `diff $_->[0] A B` }, @tests 

    for (  @diffs )
        s/(Sun|Mon|Tue|Wed|Thu|Fri|Sat).*/<MTIME_A>/m 
        s/(Sun|Mon|Tue|Wed|Thu|Fri|Sat).*/<MTIME_B>/m 
    

    $me =~ s/^(__DATA__\n).*//ms 
    open $me_fh, ">", "$^PROGRAM_NAME" or die $^OS_ERROR 
    print $me_fh,
        $me,
        "__DATA__\n",
        join $sep, @:  "$file_options\n", < @diffs
        or die $^OS_ERROR 

    close $me_fh or die $^OS_ERROR 
    #    unlink "A" or warn "$! unlinking A" ;
    #    unlink "B" or warn "$! unlinking B" ;
    exit 0 


## Ok, we're not updating, so run the tests...

my @data = split $sep, join "", @:  ~< $^DATA
close $^DATA or die $^OS_ERROR 
die "Found " . nelems(@data),
    " elements, not ", ( nelems(@tests) + 1 ),
    ", time to --update?\n"
    unless nelems(@data) == nelems(@tests) + 1 

my @file_options = @:  eval "do \{ (" . shift( @data ) . ") \}"  
die if $^EVAL_ERROR 

my ( $mtime_A, $mtime_B ) 

do
    my %o = %:  < @file_options  
    $mtime_A = %o{?MTIME_A} 
    $mtime_B = %o{?MTIME_B} 


plan tests => scalar nelems(@tests) 
for (  @tests )
    my (@:  $diff_opts, $Diff_opts ) =  $_->@ 
    my $expect = shift @data 

    $expect =~ s/<MTIME_A>/$(localtime $mtime_A)/ 
    $expect =~ s/<MTIME_B>/$(localtime $mtime_B)/ 

    my @Diff_opts = @:  eval "($Diff_opts)"  
    die if $^EVAL_ERROR 

    my $output = diff \@A, \@B, \%:  < @file_options, < @Diff_opts  
    if ( $output eq $expect )
        ok( 1 ) 
    else
        ok( 0 ) 
        warn "# diff options: $diff_opts\n" 
        warn "# my options: $Diff_opts\n" 
        ## Merge the outputs using A::D
        my @E = split m/^/, $expect 
        my @G = split m/^/, $output 
        my $w = length "Expected" 
        for (@E +@+ @G)
            s/\n/\\n/g 
            $w = length if length +> $w 
        
        my $fmt = "# \%-$($w)s \%-2s \%-$($w)s\n" 
        printf $^STDERR, $fmt, "Expected", " ", "Got" 
        print $^STDERR, "# ", "-" x ( $w * 2 + 4 ), "\n" 

        my ( $E_start, $G_start ) 
        my $print_diff = sub (@< @_)
            my (@:  $E_end, $G_end ) =  @_ 
            if ( defined $E_start || defined $G_start )
                while ( $E_start +< $E_end || $G_start +< $G_end )
                    printf $^STDERR, (
                        $fmt,
                        $E_start +< $E_end ?? @E[$E_start] !! "",
                        "!=",
                        $G_start +< $G_end ?? @G[$G_start] !! ""
                        ) 

                    ++$E_start 
                    ++$G_start 
                
                $E_start = $G_start = undef 

            
         

        my $dis = sub (@< @_)
            $E_start = @_[0] unless defined $E_start 
            $G_start = @_[1] unless defined $G_start 
         

        traverse_sequences(
            \@E, \@G,
            \%: 
            MATCH => sub (@< @_)
                $print_diff->( < @_ )
                printf $^STDERR, $fmt, @E[@_[0]], "==", @G[@_[1]]

            DISCARD_A => $dis
            DISCARD_B => $dis
            
            ) 
        $print_diff->( scalar nelems(@E), scalar nelems(@G) ) 

        print $^STDERR, "# ", "-" x ( $w * 2 + 4 ), "\n" 
        print $^STDERR, "#\n" 
    



__DATA__
FILENAME_A => "A",
MTIME_A => 1007983243,
FILENAME_B => "B",
MTIME_B => 1007983244,

----8<--------8<--------8<--------8<--------8<--------8<--------8<----
--- A	<MTIME_A>
+++ B	<MTIME_B>
@@ -2,13 +2,13 @@
 2
 3
 4
-5d
+5a
 6
 7
 8
 9
+9a
 10
 11
-11d
 12
 13
----8<--------8<--------8<--------8<--------8<--------8<--------8<----
*** A	<MTIME_A>
--- B	<MTIME_B>
***************
*** 2,14 ****
  2
  3
  4
! 5d
  6
  7
  8
  9
  10
  11
- 11d
  12
  13
--- 2,14 ----
  2
  3
  4
! 5a
  6
  7
  8
  9
+ 9a
  10
  11
  12
  13
----8<--------8<--------8<--------8<--------8<--------8<--------8<----
*** A	<MTIME_A>
--- B	<MTIME_B>
***************
*** 5 ****
! 5d
--- 5 ----
! 5a
***************
*** 9 ****
--- 10 ----
+ 9a
***************
*** 12 ****
- 11d
--- 12 ----
----8<--------8<--------8<--------8<--------8<--------8<--------8<----
--- A	<MTIME_A>
+++ B	<MTIME_B>
@@ -5 +5 @@
-5d
+5a
@@ -9,0 +10 @@
+9a
@@ -12 +12,0 @@
-11d
----8<--------8<--------8<--------8<--------8<--------8<--------8<----
5c5
< 5d
---
> 5a
9a10
> 9a
12d12
< 11d
