# $Id: 03doc.t 684 2007-09-18 20:27:03Z pajas $

##
# this test checks the DOM Document interface of XML::LibXML
# it relies on the success of t/01basic.t and t/02parse.t

# it will ONLY test the DOM capabilities as specified in DOM Level3
# XPath tests should be done in another test file

# since all tests are run on a preparsed

use Test::More


BEGIN { plan tests => 135 };
use XML::LibXML
use XML::LibXML::Common < qw(:libxml)

do
    print $^STDOUT, "# 1. Document Attributes\n"

    my $doc = XML::LibXML::Document->createDocument()
    ok($doc)
    ok( not defined $doc->encoding)
    is( $doc->version,  "1.0" )
    is( $doc->standalone, -1 )  # is the value we get for undefined,
    # actually the same as 0 but just not set.
    ok( not defined $doc->URI)  # should be set by default.
    is( $doc->compression, -1 ) # -1 indicates NO compression at all!
    # while 0 indicates just no zip compression
    # (big difference huh?)

    $doc->setEncoding( "iso-8859-1" )
    is( $doc->encoding, "iso-8859-1" )

    $doc->setVersion(12.5)
    is( $doc->version, "12.5" )

    $doc->setStandalone(1)
    is( $doc->standalone, 1 )

    $doc->setBaseURI( "localhost/here.xml" )
    is( $doc->URI, "localhost/here.xml" )

    my $doc2 = XML::LibXML::Document->createDocument("1.1", "iso-8859-2")
    is( $doc2->encoding, "iso-8859-2" )
    is( $doc2->version,  "1.1" )
    is( $doc2->standalone,  -1 )


do
    print $^STDOUT, "# 2. Creating Elements\n"
    my $doc = XML::LibXML::Document->new()
    do
        my $node = $doc->createDocumentFragment()
        ok($node)
        is($node->nodeType, XML_DOCUMENT_FRAG_NODE)
    

    do
        my $node = $doc->createElement( "foo" )
        ok($node)
        is($node->nodeType, XML_ELEMENT_NODE )
        is($node->nodeName, "foo" )
    

    do
        print $^STDOUT, "# document with encoding\n"
        my $encdoc = XML::LibXML::Document->new( "1.0" )
        $encdoc->setEncoding( "iso-8859-1" )
        do
            my $node = $encdoc->createElement( "foo" )
            ok($node)
            is($node->nodeType, XML_ELEMENT_NODE )
            is($node->nodeName, "foo" )
        

        print $^STDOUT, "# SAX style document with encoding\n"
        my $node_def = \%: 
            Name => "object"
            LocalName => "object"
            Prefix => ""
            NamespaceURI => ""
            
        do
            my $node = $encdoc->createElement( $node_def->{Name} )
            ok($node)
            is($node->nodeType, XML_ELEMENT_NODE )
            is($node->nodeName, "object" )
        
    

    do
        # namespaced element test
        my $node = $doc->createElementNS( "http://kungfoo", "foo:bar" )
        ok($node)
        is($node->nodeType, XML_ELEMENT_NODE)
        is($node->nodeName, "foo:bar")
        is($node->prefix, "foo")
        is($node->localname, "bar")
        is($node->namespaceURI, "http://kungfoo")
    

    do
        print $^STDOUT, "# bad element creation\n"
        my @badnames = @:  ";", "&", "<><", "/", "1A"

        foreach my $name (  @badnames )
            my $node = try {$doc->createElement( $name );}
            ok( not defined $node )
        

    

    do
        my $node = $doc->createTextNode( "foo" )
        ok($node)
        is($node->nodeType, XML_TEXT_NODE )
        is($node->nodeValue, "foo" )
    

    do
        my $node = $doc->createComment( "foo" )
        ok($node)
        is($node->nodeType, XML_COMMENT_NODE )
        is($node->nodeValue, "foo" )
        is($node->toString, "<!--foo-->")
    

    do
        my $node = $doc->createCDATASection( "foo" )
        ok($node)
        is($node->nodeType, XML_CDATA_SECTION_NODE )
        is($node->nodeValue, "foo" )
        is($node->toString, "<![CDATA[foo]]>")
    

    print $^STDOUT, "# 2.1 Create Attributes\n"
    do
        my $attr = $doc->createAttribute("foo", "bar")
        ok($attr)
        is($attr->nodeType, XML_ATTRIBUTE_NODE )
        is($attr->name, "foo")
        is($attr->value, "bar" )
        is($attr->hasChildNodes, 0)
        my $content = $attr->firstChild
        ok( $content )
        is( $content->nodeType, XML_TEXT_NODE )
    
    do
        print $^STDOUT, "# bad attribute creation\n"
        my @badnames = @:  ";", "&", "<><", "/", "1A"

        foreach my $name (  @badnames )
            my $node = try {$doc->createAttribute( $name, "bar" );}
            ok( not defined $node )
        

    


    do
        try {
            my $attr = $doc->createAttributeNS("http://kungfoo", "kung:foo","bar");
        }
        ok($^EVAL_ERROR)

        my $root = $doc->createElement( "foo" )
        $doc->setDocumentElement( $root )

        my $attr
        try {
            $attr = $doc->createAttributeNS("http://kungfoo", "kung:foo","bar");
        }
        ok($attr)
        is($attr->nodeName, "kung:foo")
        is($attr->name,"foo" )
        is($attr->value, "bar" )

        $attr->setValue( q(bar&amp;) )
        is($attr->getValue, q(bar&amp;) )
    
    do
        print $^STDOUT, "# bad attribute creation\n"
        my @badnames = @:  ";", "&", "<><", "/", "1A"

        foreach my $name (  @badnames )
            my $node = try {$doc->createAttributeNS( undef, $name, "bar" );}
            ok( not defined $node )
        

    

    print $^STDOUT, "# 2.2 Create PIs\n"
    do
        my $pi = $doc->createProcessingInstruction( "foo", "bar" )
        ok($pi)
        is($pi->nodeType, XML_PI_NODE)
        is($pi->nodeName, "foo")
        is($pi->textContent, "bar")
        is($pi->getData, "bar")
    

    do
        my $pi = $doc->createProcessingInstruction( "foo" )
        ok($pi)
        is($pi->nodeType, XML_PI_NODE)
        is($pi->nodeName, "foo")
        my $data = $pi->textContent
        # undef or "" depending on libxml2 version
        ok( !defined $data or length($data)==0 )
        $data = $pi->getData
        ok( !defined $data or length($data)==0 )
        $pi->setData(q(bar&amp;))
        is( $pi->getData, q(bar&amp;))
        is($pi->textContent, q(bar&amp;))
    


do
    print $^STDOUT, "# 3.  Document Manipulation\n"
    print $^STDOUT, "# 3.1 Document Elements\n"

    my $doc = XML::LibXML::Document->new()
    my $node = $doc->createElement( "foo" )
    $doc->setDocumentElement( $node )
    my $tn = $doc->documentElement
    ok($tn)
    ok($node->isSameNode($tn))

    my $node2 = $doc->createElement( "bar" )

    $doc->appendChild($node2)
    my @cn = $doc->childNodes
    is( scalar(nelems(@cn)) , 1)
    ok(@cn[0]->isSameNode($node))

    $doc->insertBefore($node2, $node)
    @cn = $doc->childNodes
    is( scalar(nelems(@cn)) , 1)
    ok(@cn[0]->isSameNode($node))

    $doc->removeChild($node)
    @cn = $doc->childNodes
    is( scalar(nelems(@cn)) , 0)

    for ( 1..2 )
        my $nodeA = $doc->createElement( "x" )
        $doc->setDocumentElement( $nodeA )
    
    ok(1) # must not segfault here :)

    $doc->setDocumentElement( $node2 )
    @cn = $doc->childNodes
    is( scalar(nelems(@cn)) , 1)
    ok(@cn[0]->isSameNode($node2))

    my $node3 = $doc->createElementNS( "http://foo", "bar" )
    ok($node3)

    print $^STDOUT, "# 3.2 Processing Instructions\n"
    do
        my $pi = $doc->createProcessingInstruction( "foo", "bar" )
        $doc->appendChild( $pi )
        @cn = $doc->childNodes
        ok( $pi->isSameNode(@cn[-1]) )
        $pi->setData( 'bar="foo"' )
        is( $pi->textContent, 'bar="foo"')
        $pi->setData( foo=>"foo" )
        is( $pi->textContent, 'foo="foo"')

    

    print $^STDOUT, "# 3.3 Comment Nodes\n"

    print $^STDOUT, "# 3.4 DTDs\n"


do
    print $^STDOUT, "# 4. Document Storing\n"
    my $parser = XML::LibXML->new
    my $doc = $parser->parse_string("<foo>bar</foo>")

    ok( $doc )

    print $^STDOUT, "# 4.1 to file handle\n"
    do
        require IO::File
        my $fh = IO::File->new()
        if ( $fh->open( "example/testrun.xml", '>' ) )
            $doc->toFH( $fh )
            $fh->close
            ok(1)
            # now parse the file to check, if succeeded
            my $tdoc = $parser->parse_file( "example/testrun.xml" )
            ok( $tdoc )
            ok( $tdoc->documentElement )
            is( $tdoc->documentElement->nodeName, "foo" )
            is( $tdoc->documentElement->textContent, "bar" )
            unlink "example/testrun.xml" 
        
    

    print $^STDOUT, "# 4.2 to named file\n"
    do
        $doc->toFile( "example/testrun.xml" )
        ok(1)
        # now parse the file to check, if succeeded
        my $tdoc = $parser->parse_file( "example/testrun.xml" )
        ok( $tdoc )
        ok( $tdoc->documentElement )
        is( $tdoc->documentElement->nodeName, "foo" )
        is( $tdoc->documentElement->textContent, "bar" )
        unlink "example/testrun.xml" 
    

    print $^STDOUT, "# 5 ELEMENT LIKE FUNCTIONS\n"
    do
        my $parser2 = XML::LibXML->new()
        my $string1 = "<A><A><B/></A><A><B/></A></A>"
        my $string2 = '<C:A xmlns:C="xml://D"><C:A><C:B/></C:A><C:A><C:B/></C:A></C:A>'
        my $string3 = '<A xmlns="xml://D"><A><B/></A><A><B/></A></A>'
        my $string4 = '<C:A><C:A><C:B/></C:A><C:A><C:B/></C:A></C:A>'
        my $string5 = '<A xmlns:C="xml://D"><C:A>foo<A/>bar</C:A><A><C:B/>X</A>baz</A>'
        do
            my $doc2 = $parser2->parse_string($string1)
            my @as   = $doc2->getElementsByTagName( "A" )
            is( scalar( nelems(@as) ), 3)

            @as   = $doc2->getElementsByTagName( "*" )
            is( scalar( nelems(@as) ), 5)

            @as   = $doc2->getElementsByTagNameNS( "*", "B" )
            is( scalar( nelems(@as) ), 2)

            @as   = $doc2->getElementsByLocalName( "A" )
            is( scalar( nelems(@as) ), 3)

            @as   = $doc2->getElementsByLocalName( "*" )
            is( scalar( nelems(@as) ), 5)
        
        do
            my $doc2 = $parser2->parse_string($string2)
            my @as   = $doc2->getElementsByTagName( "C:A" )
            is( scalar( nelems(@as) ), 3)
            @as   = $doc2->getElementsByTagNameNS( "xml://D", "A" )
            is( scalar( nelems(@as) ), 3)
            @as   = $doc2->getElementsByTagNameNS( "*", "A" )
            is( scalar( nelems(@as) ), 3)
            @as   = $doc2->getElementsByLocalName( "A" )
            is( scalar( nelems(@as) ), 3)
        
        do
            my $doc2 = $parser2->parse_string($string3)
            #            my @as   = $doc2->getElementsByTagName( "A" );
            #            is( scalar( @as ), 3);
            my @as   = $doc2->getElementsByTagNameNS( "xml://D", "A" )
            is( scalar( nelems(@as) ), 3)
            @as   = $doc2->getElementsByLocalName( "A" )
            is( scalar( nelems(@as) ), 3)
        
        do
            $parser2->recover(1)
            local $^WARN_HOOK = sub { }
            my $doc2 = $parser2->parse_string($string4)
            #            my @as   = $doc2->getElementsByTagName( "C:A" );
            #            is( scalar( @as ), 3);
            my @as   = $doc2->getElementsByLocalName( "A" )
            is( scalar( nelems(@as) ), 3)
        
        do
            my $doc2 = $parser2->parse_string($string5)
            my @as   = $doc2->getElementsByTagName( "C:A" )
            is( scalar( nelems(@as) ), 1)
            @as   = $doc2->getElementsByTagName( "A" )
            is( scalar( nelems(@as) ), 3)
            @as   = $doc2->getElementsByTagNameNS( "*", "A" )
            is( scalar( nelems(@as) ), 4)
            @as   = $doc2->getElementsByTagNameNS( "*", "*" )
            is( scalar( nelems(@as) ), 5)
            @as   = $doc2->getElementsByTagNameNS( "xml://D", "*" )
            is( scalar( nelems(@as) ), 2)

            my $A = $doc2->getDocumentElement
            @as   = $A->getChildrenByTagName( "A" )
            is( scalar( nelems(@as) ), 1)
            @as   = $A->getChildrenByTagName( "C:A" )
            is( scalar( nelems(@as) ), 1)
            @as   = $A->getChildrenByTagName( "C:B" )
            is( scalar( nelems(@as) ), 0)
            @as   = $A->getChildrenByTagName( "*" )
            is( scalar( nelems(@as) ), 2)
            @as   = $A->getChildrenByTagNameNS( "*", "A" )
            is( scalar( nelems(@as) ), 2)
            @as   = $A->getChildrenByTagNameNS( "xml://D", "*" )
            is( scalar( nelems(@as) ), 1)
            @as   = $A->getChildrenByTagNameNS( "*", "*" )
            is( scalar( nelems(@as) ), 2)
            @as   = $A->getChildrenByLocalName( "A" )
            is( scalar( nelems(@as) ), 2)
        
    

do
    print $^STDOUT, "# 5. Bug fixes (to be use with valgrind)\n"
    do
        my $doc=XML::LibXML->createDocument() # create a doc
        my $x=$doc->createPI(foo=>"bar")      # create a PI
        undef $doc                            # should not free
        undef $x                              # free the PI
        ok(1)
    
    do
        my $doc=XML::LibXML->createDocument() # create a doc
        my $x=$doc->createAttribute(foo=>"bar") # create an attribute
        undef $doc                            # should not free
        undef $x                              # free the attribute
        ok(1)
    
    do
        my $doc=XML::LibXML->createDocument() # create a doc
        my $x=$doc->createAttributeNS(undef,foo=>"bar") # create an attribute
        undef $doc                            # should not free
        undef $x                              # free the attribute
        ok(1)
    
    do
        my $doc=XML::LibXML->new->parse_string('<foo xmlns:x="http://foo.bar"/>')
        my $x=$doc->createAttributeNS('http://foo.bar','x:foo'=>"bar") # create an attribute
        undef $doc                            # should not free
        undef $x                              # free the attribute
        ok(1)
    


