# $Id: 04node.t 632 2006-11-16 21:10:49Z pajas $

##
# this test checks the DOM Node interface of XML::LibXML
# it relies on the success of t/01basic.t and t/02parse.t

# it will ONLY test the DOM capabilities as specified in DOM Level3
# XPath tests should be done in another test file

# since all tests are run on a preparsed

use Test::More

BEGIN { plan tests => 114 };
use XML::LibXML
use XML::LibXML::Common < qw(:libxml)

my $xmlstring = q{<foo>bar<foobar/><bar foo="foobar"/><!--foo--><![CDATA[&foo bar]]></foo>}

my $parser = XML::LibXML->new()
my $doc    = $parser->parse_string( $xmlstring )

print $^STDOUT, "# 1   Standalone Without NameSpaces\n\n"
print $^STDOUT, "# 1.1 Node Attributes\n"

do
    my $node = $doc->documentElement
    my $rnode

    ok($node)
    is($node->nodeType, XML_ELEMENT_NODE)
    is($node->nodeName, "foo")
    ok(not defined $node->nodeValue)
    ok($node->hasChildNodes)
    is($node->textContent, "bar&foo bar")

    do
        my @children = $node->childNodes
        is( scalar nelems(@children), 5 )
        is( @children[0]->nodeType, XML_TEXT_NODE )
        is( @children[0]->nodeValue, "bar" )
        is( @children[4]->nodeType, XML_CDATA_SECTION_NODE )
        is( @children[4]->nodeValue, "&foo bar" )

        my $fc = $node->firstChild
        ok( $fc )
        ok( $fc->isSameNode(@children[0]))
        ok( $fc->baseURI =~ m/unknown-/ )

        my $od = $fc->ownerDocument
        ok( $od )
        ok( $od->isSameNode($doc))

        my $xc = $fc->nextSibling
        ok( $xc )
        ok( $xc->isSameNode(@children[1]) )

        $fc = $node->lastChild
        ok( $fc )
        ok( $fc->isSameNode(@children[4]))

        $xc = $fc->previousSibling
        ok( $xc )
        ok( $xc->isSameNode(@children[3]) )
        $rnode = $xc

        $xc = $fc->parentNode
        ok( $xc )
        ok( $xc->isSameNode($node) )

        $xc = @children[2]
        do
            my @attributes = $xc->attributes
            is( scalar( nelems(@attributes) ), 1 )
        

        print $^STDOUT, "# 1.2 Node Cloning\n"
        do
            my $cnode  = $doc->createElement("foo")
            $cnode->setAttribute('aaa','AAA')
            $cnode->setAttributeNS('http://ns','x:bbb','BBB')
            my $c1node = $doc->createElement("bar")
            $cnode->appendChild( $c1node )

            my $xnode = $cnode->cloneNode(0)
            ok( $xnode )
            is( $xnode->nodeName, "foo" )
            ok( not $xnode->hasChildNodes )
            is( $xnode->getAttribute('aaa'),'AAA' )
            is( $xnode->getAttributeNS('http://ns','bbb'),'BBB' )

            $xnode = $cnode->cloneNode(1)
            ok( $xnode )
            is( $xnode->nodeName, "foo" )
            ok( $xnode->hasChildNodes )
            is( $xnode->getAttribute('aaa'),'AAA' )
            is( $xnode->getAttributeNS('http://ns','bbb'),'BBB' )

            my @cn = $xnode->childNodes
            ok( nelems(@cn) )
            is( scalar(nelems(@cn)), 1)
            is( @cn[0]->nodeName, "bar" )
            ok( !@cn[0]->isSameNode( $c1node ) )

            print $^STDOUT, "# clone namespaced elements\n"
            my $nsnode = $doc->createElementNS( "fooNS", "foo:bar" )

            my $cnsnode = $nsnode->cloneNode(0)
            is( $cnsnode->nodeName, "foo:bar" )
            ok( $cnsnode->localNS() )
            is( $cnsnode->namespaceURI(), 'fooNS' )

            print $^STDOUT, "# clone namespaced elements (recursive)\n"
            my $c2nsnode = $nsnode->cloneNode(1)
            is( $c2nsnode->toString(), $nsnode->toString() )
        

        print $^STDOUT, "# 1.3 Node Value\n"
        my $string2 = "<foo>bar<tag>foo</tag></foo>"
        do
            my $doc2 = $parser->parse_string( $string2 )
            my $root = $doc2->documentElement
            ok( not defined $root->nodeValue )
            is( $root->textContent, "barfoo")
        
    

    print $^STDOUT, "# 2. (Child) Node Manipulation\n"

    print $^STDOUT, "# 2.1 Valid Operations\n"

    do
        print $^STDOUT, "# 2.1.1 Single Node\n"

        my $inode = $doc->createElement("kungfoo") # already tested
        my $jnode = $doc->createElement("kungfoo")
        my $xn = $node->insertBefore($inode, $rnode)
        ok( $xn )
        ok( $xn->isSameNode($inode) )


        $node->insertBefore( $jnode, undef )
        my @ta  = $node->childNodes()
        $xn = pop @ta
        ok( $xn->isSameNode( $jnode ) )
        $jnode->unbindNode

        my @cn = $node->childNodes
        is(scalar(nelems(@cn)), 6)
        ok( @cn[3]->isSameNode($inode) )

        $xn = $node->removeChild($inode)
        ok($xn)
        ok($xn->isSameNode($inode))

        @cn = $node->childNodes
        is(scalar(nelems(@cn)), 5)
        ok( @cn[3]->isSameNode($rnode) )

        $xn = $node->appendChild($inode)
        ok($xn)
        ok($xn->isSameNode($inode))
        ok($xn->isSameNode($node->lastChild))

        $xn = $node->removeChild($inode)
        ok($xn)
        ok($xn->isSameNode($inode))
        ok(@cn[-1]->isSameNode($node->lastChild))

        $xn = $node->replaceChild( $inode, $rnode )
        ok($xn)
        ok($xn->isSameNode($rnode))

        my @cn2 = $node->childNodes
        is(scalar(nelems(@cn)), 5)
        ok( @cn2[3]->isSameNode($inode) )
    

    do
        print $^STDOUT, "\n# insertAfter Tests\n"
        my $anode = $doc->createElement("a")
        my $bnode = $doc->createElement("b")
        my $cnode = $doc->createElement("c")
        my $dnode = $doc->createElement("d")

        $anode->insertAfter( $bnode, undef )
        is( $anode->toString(), '<a><b/></a>' )

        $anode->insertAfter( $dnode, undef )
        is( $anode->toString(), '<a><b/><d/></a>' )

        $anode->insertAfter( $cnode, $bnode )
        is( $anode->toString(), '<a><b/><c/><d/></a>' )

    

    do
        print $^STDOUT, "\ntest\n" 
        my ($inode, $jnode )

        $inode = $doc->createElement("kungfoo") # already tested
        $jnode = $doc->createElement("foobar")

        my $xn = $inode->insertBefore( $jnode, undef)
        ok( $xn )
        ok( $xn->isSameNode( $jnode ) )
        print($^STDOUT,  "# ". $xn->toString() )

    

    do
        print $^STDOUT, "# 2.1.2 Document Fragment\n"

        my @cn   = $doc->documentElement->childNodes
        my $rnode= $doc->documentElement

        my $frag = $doc->createDocumentFragment
        my $node1= $doc->createElement("kung")
        my $node2= $doc->createElement("foo")

        $frag->appendChild($node1)
        $frag->appendChild($node2)

        my $xn = $node->appendChild( $frag )
        ok($xn)
        my @cn2 = $node->childNodes
        is(scalar(nelems(@cn2)), 7)
        ok(@cn2[-1]->isSameNode($node2))
        ok(@cn2[-2]->isSameNode($node1))

        $frag->appendChild( $node1 )
        $frag->appendChild( $node2 )

        @cn2 = $node->childNodes
        is(scalar(nelems(@cn2)), 5)

        $xn = $node->replaceChild( $frag, @cn[3] )
        ok($xn)
        ok($xn->isSameNode(@cn[3]))
        @cn2 = $node->childNodes
        is(scalar(nelems(@cn2)), 6)

        $frag->appendChild( $node1 )
        $frag->appendChild( $node2 )

        $xn = $node->insertBefore( $frag, @cn[0] )
        ok($xn)
        ok($node1->isSameNode($node->firstChild))
        @cn2 = $node->childNodes
        is(scalar(nelems(@cn2)), 6)
    

    print $^STDOUT, "# 2.2 Invalid Operations\n"


    print $^STDOUT, "# 2.3 DOM extensions \n"
    do
        my $str = "<foo><bar/>com</foo>"
        my $doc = XML::LibXML->new->parse_string( $str )
        my $elem= $doc->documentElement
        ok( $elem )
        ok( $elem->hasChildNodes )
        $elem->removeChildNodes
        is( $elem->hasChildNodes,0 )
        $elem->toString
    


print $^STDOUT, "# 3   Standalone With NameSpaces\n\n"

do
    my $doc = XML::LibXML::Document->new()
    my $URI ="http://kungfoo"
    my $pre = "foo"
    my $name= "bar"

    my $elem = $doc->createElementNS($URI, $pre.":".$name)

    ok($elem)
    is($elem->nodeName, $pre.":".$name)
    is($elem->namespaceURI, $URI)
    is($elem->prefix, $pre)
    is($elem->localname, $name )

    is( $elem->lookupNamespacePrefix( $URI ), $pre)
    is( $elem->lookupNamespaceURI( $pre ), $URI)

    my @ns = $elem->getNamespaces
    is( scalar(nelems(@ns)) ,1 )


print $^STDOUT, "# 4.   Document swtiching\n"

do
    print $^STDOUT, "# 4.1 simple document\n"
    my $docA = XML::LibXML::Document->new
    do
        my $docB = XML::LibXML::Document->new
        my $e1   = $docB->createElement( "A" )
        my $e2   = $docB->createElement( "B" )
        my $e3   = $docB->createElementNS( "http://kungfoo", "C:D" )
        $e1->appendChild( $e2 )
        $e1->appendChild( $e3 )

        $docA->setDocumentElement( $e1 )
    
    my $elem = $docA->documentElement
    my @c = $elem->childNodes
    my $xroot = @c[0]->ownerDocument
    ok( $xroot->isSameNode($docA) )




print $^STDOUT, "# 5.   libxml2 specials\n"

do
    my $docA = XML::LibXML::Document->new
    my $e1   = $docA->createElement( "A" )
    my $e2   = $docA->createElement( "B" )
    my $e3   = $docA->createElement( "C" )

    $e1->appendChild( $e2 )
    my $x = $e2->replaceNode( $e3 )
    my @cn = $e1->childNodes
    ok(nelems(@cn))
    is( scalar(nelems(@cn)), 1 )
    ok(@cn[0]->isSameNode($e3))
    ok($x->isSameNode($e2))

    $e3->addSibling( $e2 )
    @cn = $e1->childNodes
    is( scalar(nelems(@cn)), 2 )
    ok(@cn[0]->isSameNode($e3))
    ok(@cn[1]->isSameNode($e2))


print $^STDOUT, "# 7. importing and adopting\n"

do
    my $parser = XML::LibXML->new
    my $doc1 = $parser->parse_string( "<foo>bar<foobar/></foo>" )
    my $doc2 = XML::LibXML::Document->new

    ok( $doc1 && $doc2 )
    my $rnode1 = $doc1->documentElement
    ok( $rnode1 )
    my $rnode2 = $doc2->importNode( $rnode1 )
    ok( not $rnode2->isSameNode( $rnode1 ) ) 
    $doc2->setDocumentElement( $rnode2 )

    my $node = $rnode2->cloneNode(0)
    ok( $node )
    my $cndoc = $node->ownerDocument
    ok( $cndoc )
    ok( $cndoc->isSameNode( $doc2 ) )

    my $xnode = XML::LibXML::Element->new("test")

    my $node2 = $doc2->importNode($xnode)
    ok( $node2 )
    my $cndoc2 = $node2->ownerDocument
    ok( $cndoc2 )
    ok( $cndoc2->isSameNode( $doc2 ) )

    my $doc3 = XML::LibXML::Document->new
    my $node3 = $doc3->adoptNode( $xnode )
    ok( $node3 )
    ok( $xnode->isSameNode( $node3 ) )
    ok( $doc3->isSameNode( $node3->ownerDocument ) )

    my $xnode2 = XML::LibXML::Element->new("test")
    $xnode2->setOwnerDocument( $doc3 ) # alternate version of adopt node
    ok( $xnode2->ownerDocument )
    ok( $doc3->isSameNode( $xnode2->ownerDocument ) )


do
    # appending empty fragment
    my $doc = XML::LibXML::Document->new()
    my $frag = $doc->createDocumentFragment()
    my $root = $doc->createElement( 'foo' )
    my $r = $root->appendChild( $frag )
    ok( $r )


do
    for my $obj (qw(Document DocumentFragment Comment CDATASection PI Text)) {

    }


