# -*- cperl -*-
use Test::More
BEGIN { plan tests=>124; }
use XML::LibXML
use XML::LibXML::Common < qw(:libxml)

my $parser = XML::LibXML->new()

my $xml1 = <<EOX
<a xmlns:b="http://whatever"
><x b:href="out.xml"
/><b:c/></a>
EOX

my $xml2 = <<EOX
<a xmlns:b="http://whatever" xmlns:c="http://kungfoo"
><x b:href="out.xml"
/><b:c/><c:b/></a>
EOX

my $xml3 = <<EOX
<a xmlns:b="http://whatever">
    <x b:href="out.xml"/>
    <x>
    <c:b xmlns:c="http://kungfoo">
        <c:d/>
    </c:b>
    </x>
    <x>
    <c:b xmlns:c="http://foobar">
        <c:d/>
    </c:b>
    </x>
</a>
EOX

print $^STDOUT, "# 1.   single namespace \n"

do
    my $doc1 = $parser->parse_string( $xml1 )
    my $elem = $doc1->documentElement
    is($elem->lookupNamespaceURI( "b" ), "http://whatever" )
    my @cn = $elem->childNodes
    is(@cn[0]->lookupNamespaceURI( "b" ), "http://whatever" )
    is(@cn[1]->namespaceURI, "http://whatever" )


print $^STDOUT, "# 2.    multiple namespaces \n"

do
    my $doc2 = $parser->parse_string( $xml2 )

    my $elem = $doc2->documentElement
    is($elem->lookupNamespaceURI( "b" ), "http://whatever")
    is($elem->lookupNamespaceURI( "c" ), "http://kungfoo")
    my @cn = $elem->childNodes

    is(@cn[0]->lookupNamespaceURI( "b" ), "http://whatever" )
    is(@cn[0]->lookupNamespaceURI( "c" ), "http://kungfoo")

    is(@cn[1]->namespaceURI, "http://whatever" )
    is(@cn[2]->namespaceURI, "http://kungfoo" )


print $^STDOUT, "# 3.   nested names \n"

do
    my $doc3 = $parser->parse_string( $xml3 )
    my $elem = $doc3->documentElement
    my @cn = $elem->childNodes
    my @xs = grep { $_->nodeType == XML_ELEMENT_NODE }, @cn

    my @x1 = @xs[1]->childNodes; my @x2 = @xs[2]->childNodes

    is( @x1[1]->namespaceURI , "http://kungfoo" )
    is( @x2[1]->namespaceURI , "http://foobar" )

    # namespace scopeing
    ok( not defined $elem->lookupNamespacePrefix( "http://kungfoo" ) )
    ok( not defined $elem->lookupNamespacePrefix( "http://foobar" ) )


print $^STDOUT, "# 4. post creation namespace setting\n"
do
    my $e1 = XML::LibXML::Element->new("foo")
    my $e2 = XML::LibXML::Element->new("bar:foo")
    my $e3 = XML::LibXML::Element->new("foo")
    $e3->setAttribute( "kung", "foo" )
    my $a = $e3->getAttributeNode("kung")

    $e1->appendChild($e2)
    $e2->appendChild($e3)
    ok( $e2->setNamespace("http://kungfoo", "bar") )
    ok( $a->setNamespace("http://kungfoo", "bar") )
    is( $a->nodeName, "bar:kung" )


print $^STDOUT, "# 5. importing namespaces\n"

do

    my $doca = XML::LibXML->createDocument
    my $docb = XML::LibXML->new()->parse_string( <<EOX )
<x:a xmlns:x="http://foo.bar"><x:b/></x:a>
EOX

    my $b = $docb->documentElement->firstChild

    my $c = $doca->importNode( $b )

    my @attra = $c->attributes
    is( scalar(nelems(@attra)), 1 )
    is( @attra[0]->nodeType, 18 )
    my $d = $doca->adoptNode($b)

    ok( $d->isSameNode( $b ) )
    my @attrb = $d->attributes
    is( scalar(nelems(@attrb)), 1 )
    is( @attrb[0]->nodeType, 18 )


print $^STDOUT, "# 6. lossless setting of namespaces with setAttribute\n"
# reported by Kurt George Gjerde
do
    my $doc = XML::LibXML->createDocument
    my $root = $doc->createElementNS('http://example.com', 'document')
    $root->setAttribute('xmlns:xxx', 'http://example.com')
    $root->setAttribute('xmlns:yyy', 'http://yonder.com')
    $doc->setDocumentElement( $root )

    my $strnode = $root->toString()
    ok ( $strnode =~ m/xmlns:xxx/ and $strnode =~ m/xmlns=/ )


print $^STDOUT, "# 7. namespaced attributes\n"
do
    my $doc = XML::LibXML->new->parse_string(<<'EOF')
<test xmlns:xxx="http://example.com"/>
EOF
    my $root = $doc->getDocumentElement()
    # namespaced attributes
    $root->setAttribute('xxx:attr', 'value')
    ok ( $root->getAttributeNode('xxx:attr') )
    is ( $root->getAttribute('xxx:attr'), 'value' )
    print $^STDOUT, $root->toString(1),"\n"
    ok ( $root->getAttributeNodeNS('http://example.com','attr') )
    is ( $root->getAttributeNS('http://example.com','attr'), 'value' )
    is ( $root->getAttributeNode('xxx:attr')->getNamespaceURI(), 'http://example.com')


print $^STDOUT, "# 8. changing namespace declarations\n"
do
    my $xmlns = 'http://www.w3.org/2000/xmlns/'

    my $doc = XML::LibXML->createDocument
    my $root = $doc->createElementNS('http://example.com', 'document')
    $root->setAttributeNS($xmlns, 'xmlns:xxx', 'http://example.com')
    $root->setAttribute('xmlns:yyy', 'http://yonder.com')
    $doc->setDocumentElement( $root )

    # can we get the namespaces ?
    is ( $root->getAttribute('xmlns:xxx'), 'http://example.com')
    is ( $root->getAttributeNS($xmlns,'xmlns'), 'http://example.com' )
    is ( $root->getAttribute('xmlns:yyy'), 'http://yonder.com')
    is ( $root->lookupNamespacePrefix('http://yonder.com'), 'yyy')
    is ( $root->lookupNamespaceURI('yyy'), 'http://yonder.com')

    # can we change the namespaces ?
    ok ( $root->setAttribute('xmlns:yyy', 'http://newyonder.com') )
    is ( $root->getAttribute('xmlns:yyy'), 'http://newyonder.com')
    is ( $root->lookupNamespacePrefix('http://newyonder.com'), 'yyy')
    is ( $root->lookupNamespaceURI('yyy'), 'http://newyonder.com')

    # can we change the default namespace ?
    $root->setAttribute('xmlns', 'http://other.com' )
    is ( $root->getAttribute('xmlns'), 'http://other.com' )
    is ( $root->lookupNamespacePrefix('http://other.com'), "" )
    is ( $root->lookupNamespaceURI(''), 'http://other.com' )

    # non-existent namespaces
    is ( $root->lookupNamespaceURI('foo'), undef )
    is ( $root->lookupNamespacePrefix('foo'), undef )
    is ( $root->getAttribute('xmlns:foo'), undef )

    # changing namespace declaration URI and prefix
    ok ( $root->setNamespaceDeclURI('yyy', 'http://changed.com') )
    is ( $root->getAttribute('xmlns:yyy'), 'http://changed.com')
    is ( $root->lookupNamespaceURI('yyy'), 'http://changed.com')
    try { $root->setNamespaceDeclPrefix('yyy','xxx'); }
    ok ( $^EVAL_ERROR )  # prefix occupied
    try { $root->setNamespaceDeclPrefix('yyy',''); }
    ok ( $^EVAL_ERROR )  # prefix occupied
    ok ( $root->setNamespaceDeclPrefix('yyy', 'zzz') )
    is ( $root->lookupNamespaceURI('yyy'), undef )
    is ( $root->lookupNamespaceURI('zzz'), 'http://changed.com' )
    ok ( $root->setNamespaceDeclURI('zzz',undef ) )
    is ( $root->lookupNamespaceURI('zzz'), undef )
    my $strnode = $root->toString()
    ok ( $strnode !~ m/xmlns:zzz/ )

    # changing the default namespace declaration
    ok ( $root->setNamespaceDeclURI('','http://test') )
    is ( $root->lookupNamespaceURI(''), 'http://test' )
    is ( $root->getNamespaceURI(), 'http://test' )

    # changing prefix of the default ns declaration
    ok ( $root->setNamespaceDeclPrefix('','foo') )
    is ( $root->lookupNamespaceURI(''), undef )
    is ( $root->lookupNamespaceURI('foo'), 'http://test' )
    is ( $root->getNamespaceURI(),  'http://test' )
    is ( $root->prefix(),  'foo' )

    # turning a ns declaration to a default ns declaration
    ok ( $root->setNamespaceDeclPrefix('foo','') )
    is ( $root->lookupNamespaceURI('foo'), undef )
    is ( $root->lookupNamespaceURI(''), 'http://test' )
    is ( $root->lookupNamespaceURI(undef), 'http://test' )
    is ( $root->getNamespaceURI(),  'http://test' )
    is ( $root->prefix(),  undef )

    # removing the default ns declaration
    ok ( $root->setNamespaceDeclURI('',undef) )
    is ( $root->lookupNamespaceURI(''), undef )
    is ( $root->getNamespaceURI(), undef )

    $strnode = $root->toString()
    ok ( $strnode !~ m/xmlns=/ )

    # namespaced attributes
    $root->setAttribute('xxx:attr', 'value')
    ok ( $root->getAttributeNode('xxx:attr') )
    is ( $root->getAttribute('xxx:attr'), 'value' )
    print $^STDOUT, $root->toString(1),"\n"
    ok ( $root->getAttributeNodeNS('http://example.com','attr') )
    is ( $root->getAttributeNS('http://example.com','attr'), 'value' )
    is ( $root->getAttributeNode('xxx:attr')->getNamespaceURI(), 'http://example.com')

    # removing other xmlns declarations
    $root->addNewChild('http://example.com', 'xxx:foo')
    ok( $root->setNamespaceDeclURI('xxx',undef) )
    is ( $root->lookupNamespaceURI('xxx'), undef )
    is ( $root->getNamespaceURI(), undef )
    is ( $root->firstChild->getNamespaceURI(), undef )
    is ( $root->prefix(),  undef )
    is ( $root->firstChild->prefix(),  undef )


    print $^STDOUT, $root->toString(1),"\n"
    # check namespaced attributes
    is ( $root->getAttributeNode('xxx:attr'), undef )
    is ( $root->getAttributeNodeNS('http://example.com', 'attr'), undef )
    ok ( $root->getAttributeNode('attr') )
    is ( $root->getAttribute('attr'), 'value' )
    ok ( $root->getAttributeNodeNS(undef,'attr') )
    is ( $root->getAttributeNS(undef,'attr'), 'value' )
    is ( $root->getAttributeNode('attr')->getNamespaceURI(), undef)


    $strnode = $root->toString()
    ok ( $strnode !~ m/xmlns=/ )
    ok ( $strnode !~ m/xmlns:xxx=/ )
    ok ( $strnode =~ m/<foo/ )

    ok ( $root->setNamespaceDeclPrefix('xxx',undef) )

    is ( nelems($doc->findnodes('/document/foo')), 1 )
    is ( nelems($doc->findnodes('/document[foo]')), 1 )
    is ( nelems($doc->findnodes('/document[*]')), 1 )
    is ( nelems($doc->findnodes('/document[@attr and foo]')), 1 )
    is ( $doc->findvalue('/document/@attr'), 'value' )

    my $xp = XML::LibXML::XPathContext->new($doc)
    is ( nelems($xp->findnodes('/document/foo')), 1 )
    is ( nelems($xp->findnodes('/document[foo]')), 1 )
    is ( nelems($xp->findnodes('/document[*]')), 1 )

    is ( nelems($xp->findnodes('/document[@attr and foo]')), 1 )
    is ( $xp->findvalue('/document/@attr'), 'value' )

    is ( $root->firstChild->prefix(),  undef )


print $^STDOUT, "# 9. namespace reconciliation\n"
do
    my $doc = XML::LibXML->createDocument( 'http://default', 'root' )
    my $root = $doc->documentElement
    $root->setNamespace( 'http://children', 'child', 0 )

    $root->appendChild( my $n = $doc->createElementNS( 'http://default', 'branch' ))
    # appending an element in the same namespace will
    # strip its declaration
    ok( !defined($n->getAttribute( 'xmlns' )) )

    $n->appendChild( my $a = $doc->createElementNS( 'http://children', 'child:a' ))
    $n->appendChild( my $b = $doc->createElementNS( 'http://children', 'child:b' ))

    $n->appendChild( my $c = $doc->createElementNS( 'http://children', 'child:c' ))
    # appending $c strips the declaration
    ok( !defined($c->getAttribute('xmlns:child')) )

    # add another prefix for children
    $c->setAttribute( 'xmlns:foo', 'http://children' )
    is( $c->getAttribute( 'xmlns:foo' ), 'http://children' )

    $n->appendChild( my $d = $doc->createElementNS( 'http://other', 'branch' ))
    # appending an element with a new default namespace
    # will leave it declared
    is( $d->getAttribute( 'xmlns' ), 'http://other' )

    my $doca = XML::LibXML->createDocument( 'http://default/', 'root' )
    $doca->adoptNode( $a )
    $doca->adoptNode( $b )
    $doca->documentElement->appendChild( $a )
    $doca->documentElement->appendChild( $b )

    # Because the child namespace isn't defined in $doca
    # it should get declared on both child nodes $a and $b
    is( $a->getAttribute( 'xmlns:child' ), 'http://children' )
    is( $b->getAttribute( 'xmlns:child' ), 'http://children' )

    $doca = XML::LibXML->createDocument( 'http://children', 'child:root' )
    $doca->adoptNode( $a )
    $doca->documentElement->appendChild( $a )

    # $doca declares the child namespace, so the declaration
    # should now get stripped from $a
    ok( !defined($a->getAttribute( 'xmlns:child' )) )

    $doca->documentElement->removeChild( $a )

    # $a should now have its namespace re-declared
    is( $a->getAttribute( 'xmlns:child' ), 'http://children' )

    $doca->documentElement->appendChild( $a )

    # $doca declares the child namespace, so the declaration
    # should now get stripped from $a
    ok( !defined($a->getAttribute( 'xmlns:child' )) )


    $doc = XML::LibXML::Document->new
    $n = $doc->createElement( 'didl' )
    $n->setAttribute( "xmlns:xsi"=>"http://www.w3.org/2001/XMLSchema-instance" )

    $a = $doc->createElement( 'dc' )
    $a->setAttribute( "xmlns:xsi"=>"http://www.w3.org/2001/XMLSchema-instance" )
    $a->setAttribute( "xsi:schemaLocation"=>"http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives
.org/OAI/2.0/oai_dc.xsd" )

    $n->appendChild( $a )

    # the declaration for xsi should be stripped
    ok( !defined($a->getAttribute( 'xmlns:xsi' )) )

    $n->removeChild( $a )

    # should be a new declaration for xsi in $a
    is( $a->getAttribute( 'xmlns:xsi' ), 'http://www.w3.org/2001/XMLSchema-instance' )

    $b = $doc->createElement( 'foo' )
    $b->setAttribute( 'xsi:bar', 'bar' )
    $n->appendChild( $b )
    $n->removeChild( $b )

    # a prefix without a namespace can't be reliably compared,
    # so $b doesn't acquire a declaration from $n!
    ok( !defined($b->getAttribute( 'xmlns:xsi' )) )

    # tests for reconciliation during setAttributeNodeNS
    my $attr = $doca->createAttributeNS('http://children',
        'child:attr','value')
    ok($attr)
    my $child= $doca->documentElement->firstChild
    ok($child)
    $child->setAttributeNodeNS($attr)
    ok ( !defined($child->getAttribute( 'xmlns:child' )) )

    # due to libxml2 limitation, XML::LibXML declares the namespace
    # on the root element
    $attr = $doca->createAttributeNS('http://other','other:attr','value')
    ok($attr)
    $child->setAttributeNodeNS($attr)
    #
    ok ( !defined($child->getAttribute( 'xmlns:other' )) )
    ok ( defined($doca->documentElement->getAttribute( 'xmlns:other' )) )


print $^STDOUT, "# 10. xml namespace\n"
do
    my $docOne = XML::LibXML->new->parse_string('<foo><inc xml:id="test"/></foo>')
    my $docTwo = XML::LibXML->new->parse_string('<bar><urgh xml:id="foo"/></bar>')

    my $inc = $docOne->getElementById('test')
    my $rep = $docTwo->getElementById('foo')
    $inc->parentNode->replaceChild($rep, $inc)
    is($inc->getAttributeNS('http://www.w3.org/XML/1998/namespace','id'),'test')
    ok($inc->isSameNode($docOne->getElementById('test')))

