# $Id: 9_Strict.t,v 1.5 2003/05/18 08:43:14 grantm Exp $
# vim: syntax=perl

use Test::More

$^WARNING = 1

plan tests => 37


##############################################################################
#                      T E S T   R O U T I N E S
##############################################################################

eval "use XML::Simple < qw(:strict);"; die if $^EVAL_ERROR
ok(!$^EVAL_ERROR, 'XML::Simple loads ok with qw(:strict)')

# Check that the basic functionality still works

my $xml = q(<opt name1="value1" name2="value2"></opt>)

$^EVAL_ERROR = ''
my $opt = try {
    XMLin($xml, forcearray => 1, keyattr => $%);
}
is($^EVAL_ERROR, '', 'XMLin() did not fail')

my $keys = join(' ', sort keys $opt)

is($keys, 'name1 name2', 'and managed to produce the expected results')


# Confirm that forcearray cannot be omitted

try {
    $opt = XMLin($xml, keyattr => $%);
}

ok( $^EVAL_ERROR, 'omitting forcearray was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'forcearray'/,
     'with the correct error message')


# Confirm that keyattr cannot be omitted

try {
    $opt = XMLin($xml, forcearray => \$@);
}

ok( $^EVAL_ERROR, 'omitting keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'keyattr'/,
     'with the correct error message')


# Confirm that element names from keyattr cannot be omitted from forcearray

try {
    $opt = XMLin($xml, keyattr => (%:  part => 'partnum' ), forcearray => 0);
}

ok( $^EVAL_ERROR, 'omitting forcearray for elements in keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<part> set in keyattr but not in forcearray/,
     'with the correct error message')


dies_like(
  sub (@< @_)
      $opt = XMLin($xml, keyattr => (%:  part => 'partnum' ), forcearray => \(@: 'x','y'))
  ,
  qr/(?i)<part> set in keyattr but not in forcearray/,
  'omitting keyattr elements from forcearray was a fatal error')


# Confirm that missing key attributes are detected

$xml = q(
<opt>
  <part partnum="12345" desc="Thingy" />
  <part partnum="67890" desc="Wotsit" />
  <part desc="Fnurgle" />
</opt>
)

try {
    $opt = XMLin($xml, keyattr => (%:  part => 'partnum' ), forcearray => 1);
}

ok( $^EVAL_ERROR, 'key attribute missing from names element was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<part> element has no 'partnum' key attribute/,
     'with the correct error message')


# Confirm that stringification of references is trapped

$xml = q(
<opt>
  <item>
    <name><firstname>Bob</firstname></name>
    <age>21</age>
  </item>
</opt>
)

try {
    $opt = XMLin($xml, keyattr => (%:  item => 'name' ), forcearray => \(@: 'item'));
}

ok( $^EVAL_ERROR, 'key attribute not a scalar was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<item> element has non-scalar 'name' key attribute/,
     'with the correct error message')


##############################################################################
# Now confirm that XMLout gets checked too
#


# Check that the basic functionality still works under :strict

my $ref = %: 
    person => @: 
        %:  name => 'bob'
        %:  name => 'kate'
        
    

$^EVAL_ERROR = ''
$xml = try {
    XMLout($ref, keyattr => $%, rootname => 'list');
}
is($^EVAL_ERROR, '', 'XMLout() did not fail')

like($xml, qr{
      ^\s*<list\s*>
      \s*<person\s+name="bob"\s*/>
      \s*<person\s+name="kate"\s*/>
      \s*</list\s*>\s*$
    }xs, 'and managed to produce the expected results')


# Confirm that keyattr cannot be omitted

$^EVAL_ERROR = ''
try {
    XMLout($ref, rootname => 'list');
}

ok($^EVAL_ERROR, 'omitting keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'keyattr'/,
     'with the correct error message')


# Confirm that forcearray can be omitted (only rqd on input)

$^EVAL_ERROR = ''
try {
    XMLout($ref, keyattr => (%: x => 'y'));
}

is($^EVAL_ERROR, '', 'omitting forcearray was not a fatal error on output')


##############################################################################
# Now repeat all that using the OO syntax
##############################################################################

# Check that the basic functionality still works

$xml = q(<opt name1="value1" name2="value2"></opt>)

my $xs = XML::Simple->new(forcearray => 1, keyattr => $%)

$^EVAL_ERROR = ''
$opt = try {
    $xs->XMLin($xml);
}
is($^EVAL_ERROR, '', '$xs->XMLin() did not fail')

$keys = join(' ', sort keys $opt)

is($keys, 'name1 name2', 'and managed to produce the expected results')

# Confirm that forcearray cannot be omitted

$xs = XML::Simple->new(keyattr => $%)

$^EVAL_ERROR = ''
try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'omitting forcearray was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'forcearray'/,
     'with the correct error message')


# Confirm that keyattr cannot be omitted

$xs = XML::Simple->new(forcearray => \$@)

try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'omitting keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'keyattr'/,
     'with the correct error message')


# Confirm that element names from keyattr cannot be omitted from forcearray

$xs = XML::Simple->new(keyattr => (%:  part => 'partnum' ), forcearray => 0)

try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'omitting forcearray for elements in keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<part> set in keyattr but not in forcearray/,
     'with the correct error message')


$xs = XML::Simple->new(keyattr => (%:  part => 'partnum' ), forcearray => \(@: 'x','y'))

try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'omitting keyattr elements from forcearray was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<part> set in keyattr but not in forcearray/,
     'with the correct error message')


# Confirm that missing key attributes are detected

$xml = q(
<opt>
  <part partnum="12345" desc="Thingy" />
  <part partnum="67890" desc="Wotsit" />
  <part desc="Fnurgle" />
</opt>
)

$xs = XML::Simple->new(keyattr => (%:  part => 'partnum' ), forcearray => 1)
try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'key attribute missing from names element was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<part> element has no 'partnum' key attribute/,
     'with the correct error message')


# Confirm that stringification of references is trapped

$xml = q(
<opt>
  <item>
    <name><firstname>Bob</firstname></name>
    <age>21</age>
  </item>
</opt>
)

$xs = XML::Simple->new(keyattr => (%:  item => 'name' ), forcearray => \(@: 'item'))

try {
    $xs->XMLin($xml);
}

ok($^EVAL_ERROR, 'key attribute not a scalar was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)<item> element has non-scalar 'name' key attribute/,
     'with the correct error message')


##############################################################################
# Now confirm that XMLout gets checked too
#


# Check that the basic functionality still works under :strict

$ref = %: 
    person => @: 
        %:  name => 'bob'
        %:  name => 'kate'
        
    

$xs = XML::Simple->new(keyattr => $%, rootname => 'list')

$^EVAL_ERROR = ''
$xml = try {
    $xs->XMLout($ref);
}
is($^EVAL_ERROR, '', 'XMLout() did not fail')

like($xml, qr{
      ^\s*<list\s*>
      \s*<person\s+name="bob"\s*/>
      \s*<person\s+name="kate"\s*/>
      \s*</list\s*>\s*$
    }xs, 'and managed to produce the expected results')


# Confirm that keyattr cannot be omitted

$xs = XML::Simple->new(rootname => 'list')

try {
    $xs->XMLout($ref);
}

ok($^EVAL_ERROR, 'omitting keyattr was a fatal error')
like($^EVAL_ERROR->{?description}, qr/(?i)No value specified for 'keyattr'/,
     'with the correct error message')


exit(0)

