# $Id: AnyDBM_File.pm,v 1.11 2003/10/23 19:11:33 uid39246 Exp $

package WWW::RobotRules::AnyDBM_File

require  WWW::RobotRules
our @ISA = qw(WWW::RobotRules)
our $VERSION = sprintf("\%d.\%02d", q$Revision: 1.11 $ =~ m/(\d+)\.(\d+)/)

use Carp ()
use DB_File
use Fcntl


=head1 NAME

WWW::RobotRules::AnyDBM_File - Persistent RobotRules

=head1 SYNOPSIS

 require WWW::RobotRules::AnyDBM_File;
 require LWP::RobotUA;

 # Create a robot useragent that uses a diskcaching RobotRules
 my $rules = new WWW::RobotRules::AnyDBM_File 'my-robot/1.0', 'cachefile';
 my $ua = new WWW::RobotUA 'my-robot/1.0', 'me@foo.com', $rules;

 # Then just use $ua as usual
 $res = $ua->request($req);

=head1 DESCRIPTION

This is a subclass of I<WWW::RobotRules> that uses the AnyDBM_File
package to implement persistent diskcaching of F<robots.txt> and host
visit information.

The constructor (the new() method) takes an extra argument specifying
the name of the DBM file to use.  If the DBM file already exists, then
you can specify undef as agent name as the name can be obtained from
the DBM database.

=cut

sub new($class, $ua, $file)
    Carp::croak('WWW::RobotRules::AnyDBM_File filename required') unless $file

    my $self = bless \$%, $class
    $self->{+'filename'} = $file
    $self->{+'dbm'} = DB_File->new( $file, O_CREAT^|^O_RDWR, 0640 )
        or Carp::croak("Can't open $file: $^OS_ERROR")

    if ($ua)
        $self->agent($ua)
    else
        # Try to obtain name from DBM file
        $ua = $self->{'dbm'}->FETCH("|ua-name|")
        Carp::croak("No agent name specified") unless $ua
    

    $self


sub agent($self, ?$newname)
    my $old = $self->{'dbm'}->FETCH("|ua-name|")
    if (defined $newname)
        $newname =~ s!/?\s*\d+.\d+\s*$!!  # loose version
        unless ($old && $old eq $newname)
            # Old info is now stale.
            my $file = $self->{?'filename'}
            $self->{'dbm'}->sync
            $self->{'dbm'} = undef
            $self->{'dbm'} = DB_File->new( $file, O_TRUNC^|^O_RDWR, 0640 )
            $self->{'dbm'}->clear
            $self->{'dbm'}->put("|ua-name|" => $newname)
        
    
    $old


sub no_visits($self, $netloc)
    my $t = $self->{'dbm'}->FETCH("$netloc|vis")
    return 0 unless $t
    split(m/;\s*/, $t)[0]


sub last_visit($self, $netloc)
    my $t = $self->{'dbm'}->FETCH("$netloc|vis")
    return undef unless $t
    split(m/;\s*/, $t)[1]


sub fresh_until($self, $netloc, ?$fresh)
    my $old = $self->{'dbm'}->FETCH("$netloc|exp")
    if ($old)
        $old =~ s/;.*//  # remove cleartext
    
    if (defined $fresh)
        $fresh .= "; " . localtime($fresh)
        $self->{'dbm'}->put("$netloc|exp" => $fresh)
    
    $old


sub visit($self, $netloc, ?$time)
    $time ||= time

    my $count = 0
    my $old = $self->{'dbm'}->FETCH("$netloc|vis")
    if ($old)
        my $last
        (@: $count,$last, ...) =  split(m/;\s*/, $old)
        $time = $last if $last +> $time
    
    $count++
    $self->{'dbm'}->put("$netloc|vis" => "$count; $time; " . localtime($time) )


sub push_rules($self, $netloc, @< @rules)
    my $cnt = 1
    $cnt++ while $self->{'dbm'}->FETCH("$netloc|r$cnt")

    foreach ( @rules)
        $self->{'dbm'}->put("$netloc|r$cnt" => $_)
        $cnt++
    


sub clear_rules($self, $netloc)
    my $cnt = 1
    while ($self->{'dbm'}->FETCH("$netloc|r$cnt"))
        $self->{'dbm'}->del("$netloc|r$cnt")
        $cnt++
    


sub rules($self, $netloc)
    my @rules = $@
    my $cnt = 1
    while (1)
        my $rule = $self->{'dbm'}->FETCH("$netloc|r$cnt")
        last unless $rule
        push(@rules, $rule)
        $cnt++
    
    @rules


sub dump
    {
}

1

=head1 SEE ALSO

L<WWW::RobotRules>, L<LWP::RobotUA>

=head1 AUTHORS

Hakan Ardo E<lt>hakan@munin.ub2.lu.se>, Gisle Aas E<lt>aas@sn.no>

=cut

